#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Test suite for English language Integer and Digits classes
============================================================================

"""

from dragonfly.test.infrastructure      import RecognitionFailure
from dragonfly.test.element_testcase    import ElementTestCase
from dragonfly.language.base.integer    import Integer


#---------------------------------------------------------------------------
# Note: IntegerContent is imported locally in this file to avoid the
# language loader being initialised too early when run with Python 3.x.

class IntegerTestCase(ElementTestCase):
    """ Verify various integers between 0 and 10**12. """
    def _build_element(self):
        from dragonfly.language.en.number       import IntegerContent
        return Integer(content=IntegerContent, min=0, max=10**12 - 1)
    input_output = [
                    ("zero",       0),
                    ("oh",         0),
                    ("one",        1),
                    ("two",        2),
                    ("three",      3),
                    ("four",       4),
                    ("five",       5),
                    ("six",        6),
                    ("seven",      7),
                    ("eight",      8),
                    ("nine",       9),
                    ("ten",       10),
                    ("eleven",    11),
                    ("twelve",    12),
                    ("thirteen",  13),
                    ("fourteen",  14),
                    ("fifteen",   15),
                    ("sixteen",   16),
                    ("seventeen", 17),
                    ("eighteen",  18),
                    ("nineteen",  19),
                    ("seventy four hundred",    7400),
                    ("seventy four thousand",  74000),
                    ("two hundred and thirty four thousand five hundred sixty seven", 234567),
                   ]


class Limit3to14TestCase(ElementTestCase):
    """ Verify integer limits of range 3 -- 14. """
    def _build_element(self):
        from dragonfly.language.en.number       import IntegerContent
        return Integer(content=IntegerContent, min=3, max=14)
    input_output = [
                    ("oh",         RecognitionFailure),
                    ("zero",       RecognitionFailure),
                    ("one",        RecognitionFailure),
                    ("two",        RecognitionFailure),
                    ("three",      3),
                    ("four",       4),
                    ("five",       5),
                    ("six",        6),
                    ("seven",      7),
                    ("eight",      8),
                    ("nine",       9),
                    ("ten",       10),
                    ("eleven",    11),
                    ("twelve",    12),
                    ("thirteen",  13),
                    ("fourteen",   RecognitionFailure),
                    ("fifteen",    RecognitionFailure),
                    ("sixteen",    RecognitionFailure),
                    ("seventeen",  RecognitionFailure),
                    ("eighteen",   RecognitionFailure),
                    ("nineteen",   RecognitionFailure),
                   ]


class Limit23to47TestCase(ElementTestCase):
    """ Verify integer limits of range 23 -- 47. """
    def _build_element(self):
        from dragonfly.language.en.number       import IntegerContent
        return Integer(content=IntegerContent, min=23, max=47)
    input_output = [
                    ("twenty two",        RecognitionFailure),
                    ("twenty three",      23),
                    ("forty six",         46),
                    ("forty seven",       RecognitionFailure),
                   ]


class Limit230to350TestCase(ElementTestCase):
    """ Verify integer limits of range 230 -- 350. """
    def _build_element(self):
        from dragonfly.language.en.number       import IntegerContent
        return Integer(content=IntegerContent, min=230, max=350)
    input_output = [
                    ("two hundred twenty nine",     RecognitionFailure),
                    ("two hundred thirty",          230),
                    ("two hundred and thirty",      230),
                    ("two hundred and thirty zero", RecognitionFailure),
                    ("two hundred thirty one",      231),
                    ("two hundred and thirty one",  231),
                    ("three hundred forty nine",    349),
                    ("three hundred fifty zero",    RecognitionFailure),
                    ("three hundred fifty",         RecognitionFailure),
                   ]


class Limit351TestCase(ElementTestCase):
    """ Verify integer limits of range up to 351. """
    def _build_element(self):
        from dragonfly.language.en.number       import IntegerContent
        return Integer(content=IntegerContent, min=230, max=351)
    input_output = [
                    ("three hundred forty nine",    349),
                    ("three hundred fifty",         350),
                    ("three hundred fifty zero",    RecognitionFailure),
                    ("three hundred fifty one",     RecognitionFailure),
                   ]


class Limit352TestCase(ElementTestCase):
    """ Verify integer limits of range up to 352. """
    def _build_element(self):
        from dragonfly.language.en.number       import IntegerContent
        return Integer(content=IntegerContent, min=230, max=352)
    input_output = [
                    ("three hundred forty nine",    349),
                    ("three hundred fifty",         350),
                    ("three hundred fifty zero",    RecognitionFailure),
                    ("three hundred fifty one",     351),
                    ("three hundred fifty two",     RecognitionFailure),
                   ]


class ShortIntegerTestCase(ElementTestCase):
    """ Verify line integer class working as expected """
    def _build_element(self):
        from dragonfly.language.en.short_number       import ShortIntegerContent
        return Integer(content=ShortIntegerContent, min=0, max=10000000)
    input_output = [
                    ("one",                           1),
                    ("ten",                           10),
                    ("two two",                       22),
                    ("twenty three",                  23),
                    ("two three",                     23),
                    ("seventy",                       70),
                    ("seven zero",                    70),
                    ("hundred",                       100),
                    ("one oh three",                  103),
                    ("hundred three",                 103),
                    ("one twenty seven",              127),
                    ("one two seven",                 127),
                    ("one hundred twenty seven",      127),
                    ("two two two",                   222),
                    ("seven hundred",                 700),
                    ("thousand",                      1000),
                    ("seventeen hundred",             1700),
                    ("seventeen hundred fifty three", 1753),
                    ("seventeen fifty three",         1753),
                    ("one seven five three",          1753),
                    ("seventeen five three",          1753),
                    ("four thousand",                 4000),
                    ("ten thousand",                  10000),
                    ("ninety thousand",               90000),
                    ("four hundred thousand",         400000),
                    ("four hundred thousand and fifty", 400050),
                    ("four hundred thousand three hundred and forty two", 400342),
                    ("two hundred and thirty four thousand five hundred sixty seven", 234567),
                    ("five million two hundred and thirty four thousand five hundred sixty seven", 5234567),
                   ]



"""
Command-module loader for WSR.

This script can be used to look for Dragonfly command-modules for use with
the WSR engine. It scans the directory it's in and loads any ``_*.py`` it
finds.
"""


# TODO Have a simple GUI for pausing, resuming, cancelling and stopping
# recognition, etc

from __future__ import print_function

import logging
import os.path
import sys
import winsound

import six

from dragonfly import get_engine
from dragonfly import Grammar, MappingRule, Function, Dictation, FuncContext
from dragonfly.loader import CommandModuleDirectory
from dragonfly.log import setup_log


# --------------------------------------------------------------------------
# Set up basic logging.

setup_log()


# --------------------------------------------------------------------------
# User notification / rudimentary UI. MODIFY AS DESIRED

# For message in ('sleep', 'wake')
def notify(message):
    if message == 'sleep':
        print("Sleeping...")
        # get_engine().speak("Sleeping")
        play_sound('sleep')
    elif message == 'wake':
        print("Awake...")
        # get_engine().speak("Awake")
        play_sound('on')

sound_mappings = {
    "on":               r"C:\Windows\media\Speech On.wav",
    "sleep":            r"C:\Windows\media\Speech Sleep.wav",
    "off":              r"C:\Windows\media\Speech Off.wav",
    "what":             r"C:\Windows\media\Speech Misrecognition.wav",
    "dis":              r"C:\Windows\media\Speech Disambiguation.wav",
    "disambiguation":   r"C:\Windows\media\Speech Disambiguation.wav",
    "ack":              r"C:\Windows\media\Speech Disambiguation.wav",
}

def play_sound(sound, async_=True):
    if sound in sound_mappings: sound = sound_mappings[sound]
    winsound.PlaySound(sound, winsound.SND_FILENAME | (winsound.SND_ASYNC if async_ else 0))


# --------------------------------------------------------------------------
# Sleep/wake grammar.

sleeping = False

def load_sleep_wake_grammar(initial_awake):
    sleep_grammar = Grammar("sleep")

    def sleep(force=False):
        global sleeping
        if not sleeping or force:
            sleeping = True
            sleep_grammar.set_exclusiveness(True)
            get_engine().process_grammars_context()
        notify('sleep')

    def wake(force=False):
        global sleeping
        if sleeping or force:
            sleeping = False
            sleep_grammar.set_exclusiveness(False)
            get_engine().process_grammars_context()
        notify('wake')

    class SleepRule(MappingRule):
        mapping = {
            "start listening":  Function(wake),
            "stop listening":   Function(sleep),
            "halt listening":   Function(sleep),
        }
    sleep_grammar.add_rule(SleepRule())

    sleep_noise_rule = MappingRule(
        name = "sleep_noise_rule",
        mapping = { "<text>": Function(lambda text: False and print(text)) },
        extras = [ Dictation("text") ],
        context = FuncContext(lambda: sleeping),
    )
    sleep_grammar.add_rule(sleep_noise_rule)

    sleep_grammar.load()

    if initial_awake:
        wake(force=True)
    else:
        sleep(force=True)


# --------------------------------------------------------------------------
# Main event driving loop.

def main():
    logging.basicConfig(level=logging.INFO)

    try:
        path = os.path.dirname(__file__)
    except NameError:
        # The "__file__" name is not always available, for example
        # when this module is run from PythonWin.  In this case we
        # simply use the current working directory.
        path = os.getcwd()
        __file__ = os.path.join(path, "wsr_module_loader_plus.py")

    # Initialize and connect the engine.
    # Set any configuration options here as keyword arguments.
    engine = get_engine("sapi5inproc")
    engine.connect()

    # Load grammars.
    load_sleep_wake_grammar(True)
    directory = CommandModuleDirectory(path, excludes=[__file__])
    directory.load()

    # Define recognition callback functions.
    def on_begin():
        print("Speech start detected.")

    def on_recognition(words):
        message = u"Recognized: %s" % u" ".join(words)

        # This only seems to be an issue with Python 2.7 on Windows.
        if six.PY2:
            encoding = sys.stdout.encoding or "ascii"
            message = message.encode(encoding, errors='replace')
        print(message)

    def on_failure():
        print("Sorry, what was that?")

    # Start the engine's main recognition loop
    try:
        print("Listening...")
        engine.do_recognition(on_begin, on_recognition, on_failure)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



#
# This file is a command-module for Dragonfly.
# (c) Copyright 2008 by Christo Butcher
# Licensed under the LGPL, see <http://www.gnu.org/licenses/>
#

"""
Command-module loader for WSR
=============================

This script can be used to look Dragonfly command-modules 
for use with Window Speech Recognition.  It scans the 
directory it's in and loads any ``_*.py`` it finds.

"""

import os.path
import sys

import six

from dragonfly import get_engine
from dragonfly.loader import CommandModuleDirectory
from dragonfly.log import setup_log

#---------------------------------------------------------------------------
# Set up basic logging.

setup_log()
# logging.getLogger("compound.parse").setLevel(logging.INFO)


#---------------------------------------------------------------------------
# Main event driving loop.

def main():
    try:
        path = os.path.dirname(__file__)
    except NameError:
        # The "__file__" name is not always available, for example
        #  when this module is run from PythonWin.  In this case we
        #  simply use the current working directory.
        path = os.getcwd()
        __file__ = os.path.join(path, "dfly-loader-wsr.py")

    # Initialize and connect the engine.
    # Set any configuration options here as keyword arguments.
    engine = get_engine("sapi5inproc")
    engine.connect()

    # Load grammars.
    directory = CommandModuleDirectory(path, excludes=[__file__])
    directory.load()

    # Define recognition callback functions.
    def on_begin():
        print("Speech start detected.")

    def on_recognition(words):
        message = u"Recognized: %s" % u" ".join(words)

        # This only seems to be an issue with Python 2.7 on Windows.
        if six.PY2:
            encoding = sys.stdout.encoding or "ascii"
            message = message.encode(encoding, errors='replace')
        print(message)

    def on_failure():
        print("Sorry, what was that?")

    # Recognize from WSR in a loop.
    try:
        engine.do_recognition(on_begin, on_recognition, on_failure)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
This file implements interfaces to the X selections (clipboards):

 * Abstract BaseX11Clipboard interface to the X selections.
 * XselClipboard class implementation using the xsel program.

"""

# pylint: disable=W0622
# Suppress warnings about redefining the built-in 'format' function.

from __future__ import print_function

import locale
from subprocess import Popen, PIPE
import sys

from six import integer_types, binary_type

from .base_clipboard import BaseClipboard


#===========================================================================


class BaseX11Clipboard(BaseClipboard):
    """
    Base X11 clipboard class.
    """

    format_text      = 1
    format_unicode   = 13
    format_hdrop     = 15

    # Include formats for the standard three X selections.
    #: Format for the primary X selection.
    format_x_primary   = 0x10000

    #: Format for the secondary X selection.
    format_x_secondary = 0x10001

    #: Format for the clipboard X selection (alias of format_unicode).
    format_x_clipboard = 13

    format_names = {
        format_text:        "text",
        format_unicode:     "unicode",
        format_x_primary:   "x_primary",
        format_x_secondary: "x_secondary",
        format_hdrop:       "hdrop",
    }

    #-----------------------------------------------------------------------

    @classmethod
    def _get_x_selection(cls, format):
        raise NotImplementedError()

    @classmethod
    def _set_x_selection(cls, format, content):
        raise NotImplementedError()

    #-----------------------------------------------------------------------

    @classmethod
    def get_system_text(cls):
        try:
            return cls._get_x_selection(cls.format_x_clipboard)
        except TypeError:
            # Return the empty string if there is nothing in the selection
            #  or if its contents could not be retrieved.
            return u""

    @classmethod
    def set_system_text(cls, content):
        # If *content* is None, clear the clipboard and return early.
        if content is None:
            cls.clear_clipboard()
            return

        # Otherwise, convert *content* and set the system clipboard data.
        content = cls.convert_format_content(cls.format_unicode, content)
        cls._set_x_selection(cls.format_x_clipboard, content)

    @classmethod
    def clear_clipboard(cls):
        cls.set_system_text(u"")

    #-----------------------------------------------------------------------

    def copy_from_system(self, formats=None, clear=False):
        # pylint: disable=too-many-branches
        # Determine which formats to retrieve.
        supported_formats = (self.format_text, self.format_unicode,
                             self.format_hdrop, self.format_x_primary,
                             self.format_x_secondary)
        caller_specified_formats = bool(formats)
        if not formats:
            formats = supported_formats
        elif isinstance(formats, integer_types):
            formats = (formats,)

        # Verify that the given formats are valid.
        for format in formats:
            if not isinstance(format, integer_types):
                raise TypeError("Invalid clipboard format: %r"
                                % format)

        # Retrieve the system clipboard content.
        try:
            clipboard_text = self._get_x_selection(self.format_x_clipboard)
        except TypeError:
            clipboard_text = u""
        contents = {}

        # Populate the clipboard contents dictionary and raise errors for
        #  unavailable formats.
        for format in formats:
            err = False
            try:
                # Retrieve and use other X selections as text, if required.
                if format in (self.format_x_primary,
                              self.format_x_secondary):
                    text = self._get_x_selection(format)
                else:
                    text = clipboard_text

                # Attempt to convert and set text for this clipboard format.
                content = self.convert_format_content(format, text)
                contents[format] = content
            except (ValueError, TypeError):
                err = True

            # Do not raise errors for formats that the caller did not
            #  specify.
            if format in supported_formats and not caller_specified_formats:
                err = False

            # Always raise an error if the format is not supported.
            if format not in supported_formats:
                err = True

            # Raise an error if a specified clipboard format was not
            #  available.
            if err:
                format_repr = self.format_names.get(format, format)
                message = ("Specified clipboard format %r is not "
                           "available." % format_repr)
                raise TypeError(message)
        self._contents = contents

        # Then clear the system clipboard, if requested.
        if clear:
            self.clear_clipboard()

    def copy_to_system(self, clear=True):
        # Clear the system clipboard, if requested.
        if clear:
            self.clear_clipboard()

        # Transfer content to the X selections.
        # Text content is put on the clipboard selection, if necessary.
        text = self.get_text()
        if text is not None:
            self._set_x_selection(self.format_x_clipboard, text)

        # CF_HDROP content is put on the clipboard selection, if necessary.
        #  This might not work properly.
        elif self.has_format(self.format_hdrop):
            file_paths = self.get_format(self.format_hdrop)
            content = u"\n".join(file_paths) + u"\n"
            self._set_x_selection(self.format_x_clipboard, content)

        # Set the primary selection content, if necessary.
        if self.has_format(self.format_x_primary):
            content = self.get_format(self.format_x_primary)
            self._set_x_selection(self.format_x_primary, content)

        # Set the secondary selection content, if necessary.
        if self.has_format(self.format_x_secondary):
            content = self.get_format(self.format_x_secondary)
            self._set_x_selection(self.format_x_secondary, content)


#===========================================================================


class XselClipboard(BaseX11Clipboard):
    """
    Class for interacting with X selections (clipboards) using xsel.

    This is Dragonfly's default clipboard class on X11/Linux.

    """

    @classmethod
    def _run_command(cls, command, arguments, input_data=None):
        """
        Run a command with arguments and return the result.
        """
        arguments = [str(arg) for arg in arguments]
        full_command = [command] + arguments
        full_readable_command = ' '.join(full_command)
        cls._log.debug(full_readable_command)
        try:
            # Execute the child process, passing input data if necessary.
            encoding = locale.getpreferredencoding()
            popen_kwargs = dict(stdout=PIPE, stderr=PIPE)
            comm_kwargs = {}
            if input_data is not None:
                popen_kwargs["stdin"] = PIPE
                if not isinstance(input_data, binary_type):
                    input_data = input_data.encode(encoding)
                comm_kwargs["input"] = input_data
            p = Popen(full_command, **popen_kwargs)
            stdout, stderr = p.communicate(**comm_kwargs)

            # Decode output if it is binary.
            if isinstance(stdout, binary_type):
                stdout = stdout.decode(encoding)
            if isinstance(stderr, binary_type):
                stderr = stderr.decode(encoding)

            # Print error messages to stderr.
            if stderr:
                print(stderr, file=sys.stderr)

            # Return the process output and return code.
            return stdout, p.returncode
        except OSError as e:
            cls._log.error("Failed to execute command '%s': %s. Is "
                           "%s installed?",
                           full_readable_command, e, command)
            raise e

    @classmethod
    def _get_x_selection(cls, format):
        if format == cls.format_x_primary:
            arguments = ["--primary", "-o"]
        elif format == cls.format_x_secondary:
            arguments = ["--secondary", "-o"]
        elif format == cls.format_x_clipboard:
            arguments = ["--clipboard", "-o"]
        else:
            raise ValueError("Invalid X selection: %r" % format)

        stdout, return_code = cls._run_command("xsel", arguments)
        if return_code != 0 or not stdout:
            format_repr = cls.format_names.get(format, format)
            message = ("Specified X selection %r is not available."
                       % format_repr)
            raise TypeError(message)

        return stdout

    @classmethod
    def _set_x_selection(cls, format, content):
        if format == cls.format_x_primary:
            arguments = ["--primary", "-i"]
        elif format == cls.format_x_secondary:
            arguments = ["--secondary", "-i"]
        elif format == cls.format_x_clipboard:
            arguments = ["--clipboard", "-i"]
        else:
            raise ValueError("Invalid X selection: %r" % format)
        _, return_code = cls._run_command("xsel", arguments, content)
        if return_code != 0:
            format_repr = cls.format_names.get(format, format)
            message = ("Specified X selection %r could not be set."
                       % format_repr)
            raise TypeError(message)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
GrammarWrapperBase class
============================================================================

"""

try:
    from inspect import getfullargspec as getargspec
except ImportError:
    # Fallback on the deprecated function.
    from inspect import getargspec


class GrammarWrapperBase(object):

    def __init__(self, grammar, engine, recobs_manager):
        self.grammar = grammar
        self.engine = engine
        self.recobs_manager = recobs_manager

    def _process_grammar_callback(self, func, **kwargs):
        if not func:
            return

        # Only send keyword arguments that the given function accepts.
        argspec = getargspec(func)
        arg_names, kwargs_names = argspec[0], argspec[2]
        if not kwargs_names:
            kwargs = { k: v for (k, v) in kwargs.items() if k in arg_names }

        return func(**kwargs)





from dragonfly import BringApp, StartApp

#StartApp(r"c:\Windows\system32\notepad.exe").execute()
#StartApp(r"%SystemRoot%\system32\notepad.exe").execute()
#

#BringApp(r"c:\Windows\system32\notepad.exe").execute()
BringApp("winword").execute()



# This file is part of Aenea
#
# Aenea is free software: you can redistribute it and/or modify it under
# the terms of version 3 of the GNU Lesser General Public License as
# published by the Free Software Foundation.
#
# Aenea is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Aenea.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (2014) Alex Roper
# Alex Roper <alex@aroper.net>

# Heavily modified from Aenea's xdotool implementation for X11.

"""
Window class for X11
============================================================================

"""

# pylint: disable=W0622
# Suppress warnings about redefining the built-in 'id' function.

from __future__ import print_function

import locale
import logging
import os
from subprocess import Popen, PIPE
import sys

import psutil
from six import binary_type

from .base_window import BaseWindow
from .rectangle import Rectangle


class X11Window(BaseWindow):
    """
        The Window class is an interface to the window control and
        placement APIs for X11.

        Window control methods such as :meth:`close` will return ``True``
        if successful.

        This class requires the following external programs:

        * ``wmctrl``
        * ``xdotool``
        * ``xprop``

    """

    _log = logging.getLogger("window")

    #-----------------------------------------------------------------------
    # Methods and attributes for running commands.

    # Commands
    wmctrl = "wmctrl"
    xdotool = "xdotool"
    xprop = "xprop"

    @classmethod
    def _run_command(cls, command, arguments):
        """
        Run a command with arguments and return the result.

        :param command: command to run
        :type command: str
        :param arguments: arguments to append
        :type arguments: list
        :returns: stdout, return_code
        :rtype: tuple
        """
        arguments = [str(arg) for arg in arguments]
        full_command = [command] + arguments
        full_readable_command = ' '.join(full_command)
        cls._log.debug(full_readable_command)
        try:
            kwargs = dict(stdout=PIPE, stderr=PIPE)

            # Fork the process with setsid() if on a POSIX system such as
            # Linux.
            if os.name == 'posix':
                kwargs.update(dict(preexec_fn=os.setsid))

            # Execute the child process.
            p = Popen(full_command, **kwargs)
            stdout, stderr = p.communicate()

            # Decode output if it is binary.
            encoding = locale.getpreferredencoding()
            if isinstance(stdout, binary_type):
                stdout = stdout.decode(encoding)
            if isinstance(stderr, binary_type):
                stderr = stderr.decode(encoding)

            # Print error messages to stderr. Filter BadWindow messages.
            stderr = stderr.rstrip()
            if stderr and "BadWindow" not in stderr:
                print(stderr, file=sys.stderr)

            # Return the process output and return code.
            return stdout.rstrip(), p.returncode
        except OSError as e:
            cls._log.error("Failed to execute command '%s': %s. Is "
                           "%s installed?",
                           full_readable_command, e, command)
            raise e

    @classmethod
    def _run_command_simple(cls, exe, arguments):
        # Run the command and return whether or not it succeeded based on
        # the return code.
        stdout, return_code = cls._run_command(exe, arguments)
        if stdout: print(stdout)
        return return_code == 0

    @classmethod
    def _run_wmctrl_command_simple(cls, arguments):
        return cls._run_command_simple(cls.wmctrl, arguments)

    @classmethod
    def _run_xdotool_command(cls, arguments):
        return cls._run_command(cls.xdotool, arguments)

    @classmethod
    def _run_xdotool_command_simple(cls, arguments):
        return cls._run_command_simple(cls.xdotool, arguments)

    @classmethod
    def _run_xprop_command(cls, arguments):
        return cls._run_command(cls.xprop, arguments)

    #-----------------------------------------------------------------------
    # Class methods to create new Window objects.

    @classmethod
    def get_foreground(cls):
        window_id, return_code = cls._run_xdotool_command([
            "getactivewindow"
        ])
        if return_code == 0:
            return cls.get_window(int(window_id))
        else:
            return cls.get_window(0)  # return an invalid window

    @classmethod
    def get_all_windows(cls):
        # Get all window IDs using 'xdotool search'.
        stdout, return_code = cls._run_xdotool_command([
            'search', '--onlyvisible', '--name', ''
        ])
        if return_code == 0:
            lines = [line for line in stdout.split('\n') if line]
            windows = [cls.get_window(int(line)) for line in lines]
        else:
            # Return any windows found previously.
            if stdout: print(stdout)
            return list(cls._windows_by_id.values())

        # Exclude window IDs that have no associated process ID.
        result = []
        for window in windows:
            props = cls._get_properties_from_xprop(window, '_NET_WM_PID',
                                                   '_NET_WM_STATE')
            if '_NET_WM_PID' not in props:
                continue
            has_state = '_NET_WM_STATE' not in props
            result.append((window, has_state))

        # Sort the list so that windows without _NET_WM_STATE are
        # last.
        result.sort(key=lambda pair: pair[1])
        return [w for (w, _) in result]  # return just the windows

    @classmethod
    def get_matching_windows(cls, executable=None, title=None):
        # Make window searches case-insensitive.
        if executable:
            executable = executable.lower()
        if title:
            title = title.lower()

        # Get matching window IDs using 'xdotool search'.
        args = ['search', '--onlyvisible', '--name']
        if title:
            args.append(title)
        else:
            args.append('')
        stdout, return_code = cls._run_xdotool_command(args)
        if return_code == 0:
            lines = [line for line in stdout.split('\n') if line]
            windows = [cls.get_window(int(line)) for line in lines]
        else:
            # Use windows found previously.
            if stdout: print(stdout)
            windows = list(cls._windows_by_id.values())

        matching = []
        for window in windows:
            if executable:
                if window.executable.lower().find(executable) == -1:
                    continue
            if title:
                if window.title.lower().find(title) == -1:
                    continue

            # Match found.
            matching.append(window)

        # Sort the window list so that windows without _NET_WM_STATE are
        # last.
        matching.sort(key=lambda w: w.state is None)
        return matching

    #-----------------------------------------------------------------------
    # Methods for initialization and introspection.

    def __init__(self, id):
        super(X11Window, self).__init__(id=id)
        self._pid = -1  # initialized later if required
        self._executable = -1

    def __repr__(self):
        args = ["id=%d" % self._id] + list(self._names)
        return "%s(%s)" % (self.__class__.__name__, ", ".join(args))

    #-----------------------------------------------------------------------
    # Methods and properties for window attributes.

    def _get_properties_from_xprop(self, *properties):
        # This method retrieves windows properties by shelling out to xprop.
        result = {}
        args = ['-id', self.id] + list(properties)
        stdout, return_code = self._run_xprop_command(args)
        if return_code > 0:
            return {}

        for line in stdout.split('\n'):
            line = line.split(' =', 1)
            if len(line) != 2:
                continue

            raw_key, value = line
            raw_key, value = raw_key.strip(), value.strip()
            if 'STRING)' in raw_key:  # This also handles (UTF8_STRING).
                value = value[1:-1]

                # Use a list if there are multiple strings.
                if '", "' in value:
                    value = [string for string in value.split('", "')]

            # Get the key without the property type.
            key = raw_key[0:raw_key.find('(')]
            result[key] = value

        # Split up class and class name.
        if 'WM_CLASS' in result:
            window_class_name, window_class = result.pop('WM_CLASS')
            result['cls_name'] = window_class_name
            result['cls'] = window_class

        # Return the requested properties.
        return result

    def _get_window_text(self):
        # Get the title text.
        args = ['getwindowname', self.id]
        stdout, return_code = self._run_xdotool_command(args)
        if return_code == 0:
            return stdout
        else:
            if stdout: print(stdout)
            return ""

    def _get_class_name(self):
        return (self._get_properties_from_xprop("WM_CLASS")
                .get('cls_name', ''))

    def _get_window_pid(self):
        # Set the pid once when it is needed.
        if self._pid == -1:
            p = '_NET_WM_PID'
            pid = self._get_properties_from_xprop(p).get(p)
            if pid:
                pid = int(pid)
            self._pid = pid

        return self._pid

    @property
    def cls(self):
        """ Read-only access to the window's class. """
        return (self._get_properties_from_xprop("WM_CLASS")
                .get('cls', ''))

    @property
    def role(self):
        """
        Read-only access to the window's X11 role attribute.

        :returns: role
        :rtype: str
        """
        p = 'WM_WINDOW_ROLE'
        return self._get_properties_from_xprop(p).get(p, '')

    @property
    def type(self):
        """
        Read-only access to the window's X11 type property, if it is set.

        :returns: type
        :rtype: str
        """
        p = '_NET_WM_WINDOW_TYPE'
        return self._get_properties_from_xprop(p).get(p, '')

    @property
    def state(self):
        """
        Read-only access to the X window state.

        Windows can have multiple states, so this returns a tuple.

        This property invokes a (relatively) long-running function, so
        store the result locally instead of using it multiple times.

        If the window does not have the _NET_WM_STATE property, then
        ``None`` will be returned.

        :return: window state (if any)
        :rtype: tuple | None
        """
        p = '_NET_WM_STATE'
        net_wm_state = self._get_properties_from_xprop(p).get(p)
        if net_wm_state is None:
            # Indicate to callers that _NET_WM_STATE was missing.
            return None
        elif not net_wm_state:
            return tuple()
        else:
            return tuple(net_wm_state.split(', '))

    @property
    def _no_window_state(self):
        return self.state is None

    def _get_window_module(self):
        # Get the executable using the process ID and psutil.
        pid = self.pid
        if pid == -1:
            self._executable = ''
        elif self._executable == -1:
            for p in psutil.process_iter(attrs=['pid', 'exe', 'name']):
                if p.info['pid'] == pid:
                    self._executable = p.info['exe'] or p.info['name']
                    return self._executable

            # Set to '' if it wasn't found.
            self._executable = ''

        return self._executable

    @classmethod
    def _is_minimized(cls, state):
        return state is not None and '_NET_WM_STATE_HIDDEN' in state

    @property
    def is_minimized(self):
        return self._is_minimized(self.state)

    @property
    def is_visible(self):
        state = self.state
        return state is not None and '_NET_WM_STATE_HIDDEN' not in state

    @classmethod
    def _is_maximized(cls, state):
        # Note: this means a window must be maximized both horizontally and
        # vertically, but that is typically what maximize means anyway.
        return (state is not None and
                '_NET_WM_STATE_MAXIMIZED_VERT' in state and
                '_NET_WM_STATE_MAXIMIZED_HORZ' in state)

    @property
    def is_maximized(self):
        return self._is_maximized(self.state)

    @property
    def is_fullscreen(self):
        """
        Whether the window is in fullscreen mode.

        This does not work for all window types (e.g. pop up menus).

        :rtype: bool
        """
        state = self.state
        return state is not None and '_NET_WM_STATE_FULLSCREEN' in state

    @property
    def is_focused(self):
        """
        Whether the window has input focus.

        This does not work for all window types (e.g. pop up menus).

        :rtype: bool
        """
        state = self.state
        return state is not None and '_NET_WM_STATE_FOCUSED' in state

    #-----------------------------------------------------------------------
    # Methods related to window geometry.

    def get_position(self):
        stdout, return_code = self._run_xdotool_command(
            ['getwindowgeometry', '--shell', self.id])
        if return_code > 0:
            if stdout: print(stdout)
            return Rectangle(0, 0, 0, 0)

        geometry = stdout.strip().split('\n')
        geo = dict([val.lower()
                    for val in line.split('=')]
                   for line in geometry)
        geo = dict((key, int(value)) for (key, value) in geo.items())
        return Rectangle(geo['x'], geo['y'], geo['width'], geo['height'])

    def set_position(self, rectangle):
        l, t, w, h = rectangle.ltwh
        id = self.id
        return self._run_xdotool_command_simple(
            ['windowmove', id, l, t,
             'windowsize', id, w, h]
        )

    #-----------------------------------------------------------------------
    # Methods for miscellaneous window control.

    def minimize(self):
        # Attempt to minimize the window. Return the command's success.
        return self._run_xdotool_command_simple(['windowminimize',
                                                 self.id])

    def _toggle_maximize(self, is_maximized):
        # Use wmctrl to add or remove the maximized window properties from
        # the window's _NET_WM_STATE set.
        # Note: this should be possible with xprop, but is not supported.
        maximized_props = 'maximized_vert,maximized_horz'
        add_remove = 'remove' if is_maximized else 'add'
        return self._run_wmctrl_command_simple([
            '-ir', self.id, '-b', '%s,%s' % (add_remove, maximized_props)
        ])

    def maximize(self):
        if not self.is_maximized:
            return self._toggle_maximize(False)
        return True  # already maximized.

    def restore(self):
        state = self.state
        if self._is_minimized(state):
            return self._run_xdotool_command_simple([
                'windowactivate', self.id
            ])
        elif self._is_maximized(state):
            return self._toggle_maximize(True)
        else:
            # True if already restored or False if no _NET_WM_STATE.
            return state is not None

    def close(self):
        return self._run_xdotool_command_simple(['windowclose', self.id])

    def set_foreground(self):
        # Restore if minimized.
        if self.is_minimized and not self.restore():
            return False  # restore() failed
        if not self.is_focused:
            id = '%i' % self.id
            return self._run_xdotool_command_simple([
                'windowactivate', id, 'windowfocus', id
            ])

        return True

    def set_focus(self):
        """
        Set the input focus to this window.

        This method will set the input focus, but will not necessarily bring
        the window to the front.
        """
        return self._run_xdotool_command_simple([
            'windowfocus', self.id
        ])



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#


import unittest
import time
import logging
from dragonfly.engines import get_engine


#===========================================================================

class CapturingHandler(logging.Handler):

    def __init__(self):
        logging.Handler.__init__(self)
        self.records = []

    def emit(self, record):
        self.records.append(record)


#===========================================================================

class TestTimer(unittest.TestCase):

    def setUp(self):
        self.log_capture = CapturingHandler()
        logging.getLogger("engine.timer").addHandler(self.log_capture)
        self.engine = get_engine()

    def test_timer_callback_exception(self):
        """ Test handling of exceptions during timer callback. """

        callback_called = [0]
        def callback():
            callback_called[0] += 1
            raise Exception("Exception from timer callback")

        interval = 0.01
        timer = self.engine.create_timer(callback, interval)
        time.sleep(0.02)
        timer.manager.main_callback()

        # Callback was called once.
        try:
            self.assertTrue(callback_called[0] == 1)
            self.assertTrue(len(self.log_capture.records) == 1)
            log_message = self.log_capture.records[0].msg
            self.assertTrue("Exception from timer callback" in log_message)
        finally:
            # Stop the timer at the end regardless of the result.
            timer.stop()

    def test_non_repeating_timers(self):
        """ Test that non-repeating timers only run once. """

        callback_called = [0]
        def callback():
            callback_called[0] += 1

        interval = 0.01
        timer = self.engine.create_timer(callback, interval, False)
        time.sleep(0.02)
        timer.manager.main_callback()
        time.sleep(0.02)
        timer.manager.main_callback()

        # Callback was only called once.
        try:
            self.assertEqual(callback_called[0], 1)
        finally:
            # Stop the timer at the end regardless of the result.
            timer.stop()

#===========================================================================

if __name__ == "__main__":
    # Use the "text" engine by default and disable timer manager callbacks
    # to avoid race conditions.
    get_engine("text")._timer_manager.disable()

    from dragonfly.log import setup_log
    setup_log()  # tests require sane logging levels
    unittest.main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#
#   Use "1" instead of "een"
#   because the latter has
#   multiple pronunciations.

"""
Indonesian (id) language implementations of Integer and Digits classes
============================================================================

"""

from ..base.integer_internal  import (MapIntBuilder, CollectionIntBuilder,
                                      MagnitudeIntBuilder, IntegerContentBase)
from ..base.digits_internal   import DigitsContentBase


#---------------------------------------------------------------------------

int_0           = MapIntBuilder({
                                 "nul": 0
                               })
int_1_9         = MapIntBuilder({
                                 "satu":      1,
                                 "dua":       2,
                                 "tiga":      3,
                                 "empat":     4,
                                 "lima":      5,
                                 "enam":      6,
                                 "tujuh":     7,
                                 "delapan":   8,
                                 "sembilan":  9,
                               })
int_10_19       = MapIntBuilder({
                                 "sepuluh":       10,
                                 "sebelas":       11,
                                 "duabelas":      12,
                                 "tigabelas":     13,
                                 "empatbelas":    14,
                                 "limabelas":     15,
                                 "enambelas":     16,
                                 "tujuhbelas":    17,
                                 "delapanbelas":  18,
                                 "sembilanbelas": 19,
                               })
int_20_90_10    = MapIntBuilder({
                                 "duapuluh":       2,
                                 "tigapuluh":      3,
                                 "empatpuluh":     4,
                                 "limapuluh":      5,
                                 "enampuluh":      6,
                                 "tujuhpuluh":     7,
                                 "delapanpuluh":   8,
                                 "sembilanpuluh":  9,
                               })
int_20_99       = MagnitudeIntBuilder(
                   factor      = 10,
                   spec        = "[<remainder> en] <multiplier>",
                   multipliers = [int_20_90_10],
                   remainders  = [int_1_9]
                  )
int_en_1_99     = CollectionIntBuilder(
                   spec        = "[en] <element>",
                   set         = [int_1_9, int_10_19, int_20_99]
                  )
int_100s        = MagnitudeIntBuilder(
                   factor      = 100,
                   spec        = "[<multiplier>] honderd [<remainder>]",
                   multipliers = [int_1_9],
                   remainders  = [int_en_1_99]
                  )
int_100big      = MagnitudeIntBuilder(
                   factor      = 100,
                   spec        = "[<multiplier>] honderd [<remainder>]",
                   multipliers = [int_10_19, int_20_99],
                   remainders  = [int_en_1_99]
                  )
int_1000s       = MagnitudeIntBuilder(
                   factor      = 1000,
                   spec        = "[<multiplier>] duizend [<remainder>]",
                   multipliers = [int_1_9, int_10_19, int_20_99, int_100s],
                   remainders  = [int_en_1_99, int_100s]
                  )
int_1000000s    = MagnitudeIntBuilder(
                   factor      = 1000000,
                   spec        = "[<multiplier>] millioen [<remainder>]",
                   multipliers = [int_1_9, int_10_19, int_20_99, int_100s, int_1000s],
                   remainders  = [int_en_1_99, int_100s, int_1000s]
                  )


#---------------------------------------------------------------------------

class IntegerContent(IntegerContentBase):
    builders = [int_0, int_1_9, int_10_19, int_20_99,
                int_100s, int_100big, int_1000s, int_1000000s]

class DigitsContent(DigitsContentBase):
    digits = ["nul","satu","dua","tiga","empat","lima",
	          "enam","tujuh","delapan","sembilan"]




#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Window class for Windows
============================================================================

"""

# pylint: disable=E0401
# This file imports Win32-only symbols.

# pylint: disable=E0213,E1120,W0212
# Suppress warnings about _win32gui_func() lambdas and protected access.

from ctypes          import windll, pointer, c_wchar, c_ulong

import win32api
import win32con
import win32gui

from .base_window    import BaseWindow
from .rectangle      import Rectangle
from ..actions.action_key import Key


#===========================================================================

class Win32Window(BaseWindow):
    """
        The Window class is an interface to the Win32 window control
        and placement.

    """

    _results_box_class_names = ["#32770", "DgnResultsBoxWindow"]

    #-----------------------------------------------------------------------
    # Class methods to create new Window objects.

    @classmethod
    def get_foreground(cls):
        handle = win32gui.GetForegroundWindow()
        if handle in cls._windows_by_id:
            return cls._windows_by_id[handle]
        window = Win32Window(handle=handle)
        return window

    @classmethod
    def get_all_windows(cls):
        def function(handle, argument):
            argument.append(cls.get_window(handle))
        argument = []
        win32gui.EnumWindows(function, argument)
        return argument

    @classmethod
    def get_matching_windows(cls, executable=None, title=None):
        # Make window searches case-insensitive.
        if executable:
            executable = executable.lower()
        if title:
            title = title.lower()

        matching = []
        for window in cls.get_all_windows():
            if not window.is_visible:
                continue

            if executable:
                if window.executable.lower().find(executable) == -1:
                    continue
            if title:
                if window.title.lower().find(title) == -1:
                    continue

            if (window.executable.endswith("natspeak.exe")
                    and window.classname in cls._results_box_class_names
                    and window.get_position().dy < 50):
                # If a window matches the above, it is very probably
                #  the results-box of DNS.  We ignore this because
                #  its title is the words of the last recognition,
                #  which will often interfere with a search for
                #  a window with a spoken title.
                continue

            matching.append(window)

        return matching

#    @classmethod
#    def get_window_by_executable(cls, executable):
#        def function(handle, argument):
#            title = windll.user32.GetWindowText(handle)
#            print "title: %r" % title
#        windll.user32.EnumWindows(function, argument)


    #-----------------------------------------------------------------------
    # Methods for initialization and introspection.

    def __init__(self, handle):
        super(Win32Window, self).__init__(id=handle)

    def __repr__(self):
        args = ["handle=%d" % self.handle] + list(self._names)
        return "%s(%s)" % (self.__class__.__name__, ", ".join(args))

    #-----------------------------------------------------------------------
    # Direct access to various Win32 methods.

    def _win32gui_func(name):
        func = getattr(win32gui, name)
        return lambda self: func(self._handle)

    _get_rect           = _win32gui_func("GetWindowRect")
    _destroy            = _win32gui_func("DestroyWindow")
    _set_foreground     = _win32gui_func("SetForegroundWindow")
    _bring_to_top       = _win32gui_func("BringWindowToTop")
    _get_window_text    = _win32gui_func("GetWindowText")
    _get_class_name     = _win32gui_func("GetClassName")

    def _win32gui_test(name):
        test = getattr(win32gui, name)
        fget = lambda self: test(self._handle) and True or False
        return property(fget=fget,
                        doc="Shortcut to win32gui.%s() function." % name)

    is_valid        = _win32gui_test("IsWindow")
    is_enabled      = _win32gui_test("IsWindowEnabled")
    is_visible      = _win32gui_test("IsWindowVisible")
    is_minimized    = _win32gui_test("IsIconic")

    @property
    def is_maximized(self):
        # IsZoomed() is not available from win32gui for some reason.
        # So we use the function directly.
        return bool(windll.user32.IsZoomed(self._handle))

    def _win32gui_show_window(state):
        return lambda self: win32gui.ShowWindow(self._handle, state)

    minimize        = _win32gui_show_window(win32con.SW_MINIMIZE)
    maximize        = _win32gui_show_window(win32con.SW_MAXIMIZE)
    restore         = _win32gui_show_window(win32con.SW_RESTORE)

    def _win32gui_post(message, w=0, l=0):
        return lambda self: win32gui.PostMessage(self._handle, message, w, l)

    close = _win32gui_post(win32con.WM_CLOSE)

    def _get_window_pid(self):
        # Get this window's process ID.
        pid = c_ulong()
        windll.user32.GetWindowThreadProcessId(self._handle, pointer(pid))
        return pid.value

    def _get_window_module(self):
        # Get this window's process ID.
        pid = self._get_window_pid()

        # Get the process handle of this window's process ID.
        #  Access permission flags:
        #  0x0410 = PROCESS_QUERY_INFORMATION | PROCESS_VM_READ
        handle = windll.kernel32.OpenProcess(0x0410, 0, pid)

        # Retrieve and return the process's executable path.
        try:
            # Try to use the QueryForProcessImageNameW function
            #  available since Windows Vista.
            buffer_len = c_ulong(256)
            buffer = (c_wchar * buffer_len.value)()
            windll.kernel32.QueryFullProcessImageNameW(handle, 0,
                                                       pointer(buffer),
                                                       pointer(buffer_len))
            buffer = buffer[:]
            buffer = buffer[:buffer.index("\0")]
        except Exception:
            # If the function above failed, fall back to the older
            #  GetModuleFileNameEx function, available since windows XP.
            #  Note that this fallback function seems to fail when
            #  this process is 32 bit Python and handle refers to a
            #  64-bit process.
            buffer_len = 256
            buffer = (c_wchar * buffer_len)()
            windll.psapi.GetModuleFileNameExW(handle, 0, pointer(buffer),
                                              buffer_len)
            buffer = buffer[:]
            buffer = buffer[:buffer.index("\0")]
        finally:
            windll.kernel32.CloseHandle(handle)

        return str(buffer)

    #-----------------------------------------------------------------------
    # Methods related to window geometry.

    def get_position(self):
        l, t, r, b = self._get_rect()
        w = r - l; h = b - t
        return Rectangle(l, t, w, h)

    def set_position(self, rectangle):
        assert isinstance(rectangle, Rectangle)
        l, t, w, h = rectangle.ltwh
        win32gui.MoveWindow(self._handle, l, t, w, h, 1)

    #-----------------------------------------------------------------------
    # Methods for miscellaneous window control.

    def set_foreground(self):
        # Bring this window into the foreground if it isn't already the
        # current foreground window.
        if self.handle != win32gui.GetForegroundWindow():
            if self.is_minimized:
                self.restore()

            # Press a key so Windows allows us to use SetForegroundWindow()
            # (received last input event). See Microsoft's documentation on
            # SetForegroundWindow() for why this works.
            # Only do this if neither the left or right control keys are
            # held down.
            if win32api.GetKeyState(win32con.VK_CONTROL) >= 0:
                Key("control:down,control:up").execute()

            # Set the foreground window.
            self._set_foreground()

    def set_focus(self):
        # Setting window focus without raising the window doesn't appear to
        # be possible in Windows, so fallback on set_foreground().
        self.set_foreground()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
This file implements the a keyboard interface using the *pynput* Python
package.

Currently, this interface can be used on macOS (Darwin) or X11.

"""

import time

from pynput.keyboard import Controller, KeyCode, Key

from ._base import BaseKeyboard, BaseTypeable, BaseKeySymbols


class PynputTypeable(BaseTypeable):
    """ Typeable class for pynput. """

    def __init__(self, code, modifiers=(), name=None, is_text=False):
        BaseTypeable.__init__(self, code, modifiers, name, is_text)


class SafeKeyCode(object):
    """
    Class to safely get key codes from pynput.
    """

    def __getattr__(self, name):
        # Get the key code from pynput, returning KeyCode(vk=-1, char=name)
        #  if the key name isn't present.  Keys are undefined on some
        #  platforms, e.g. "pause" on Darwin.
        return getattr(Key, name, KeyCode(vk=-1, char=name))


virtual_keys = SafeKeyCode()


class BasePynputKeySymbols(BaseKeySymbols):
    """ Base key symbols for pynput. """

    # Whitespace and editing keys
    RETURN = virtual_keys.enter
    TAB = virtual_keys.tab
    SPACE = virtual_keys.space
    BACK = virtual_keys.backspace
    DELETE = virtual_keys.delete

    # Main modifier keys
    SHIFT = virtual_keys.shift
    CONTROL = virtual_keys.ctrl
    ALT = virtual_keys.alt

    # Right modifier keys
    RSHIFT = virtual_keys.shift_r
    RCONTROL = virtual_keys.ctrl_r
    RALT = virtual_keys.alt_r

    # Special keys
    ESCAPE = virtual_keys.esc
    INSERT = virtual_keys.insert
    PAUSE = virtual_keys.pause
    LSUPER = virtual_keys.cmd_l
    RSUPER = virtual_keys.cmd_r
    APPS = virtual_keys.menu
    SNAPSHOT = virtual_keys.print_screen

    # Lock keys
    SCROLL_LOCK = virtual_keys.scroll_lock
    NUM_LOCK = virtual_keys.num_lock
    CAPS_LOCK = virtual_keys.caps_lock

    # Navigation keys
    UP = virtual_keys.up
    DOWN = virtual_keys.down
    LEFT = virtual_keys.left
    RIGHT = virtual_keys.right
    PAGE_UP = virtual_keys.page_up
    PAGE_DOWN = virtual_keys.page_down
    HOME = virtual_keys.home
    END = virtual_keys.end

    # Number pad keys
    # pynput currently only exposes these for Windows, so we'll map them to
    # equivalent characters and numbers instead.
    MULTIPLY = KeyCode(char="*")
    ADD = KeyCode(char="+")
    SEPARATOR = KeyCode(char=".")  # this is locale-dependent.
    SUBTRACT = KeyCode(char="-")
    DECIMAL = KeyCode(char=".")
    DIVIDE = KeyCode(char="/")
    NUMPAD0 = KeyCode(char="0")
    NUMPAD1 = KeyCode(char="1")
    NUMPAD2 = KeyCode(char="2")
    NUMPAD3 = KeyCode(char="3")
    NUMPAD4 = KeyCode(char="4")
    NUMPAD5 = KeyCode(char="5")
    NUMPAD6 = KeyCode(char="6")
    NUMPAD7 = KeyCode(char="7")
    NUMPAD8 = KeyCode(char="8")
    NUMPAD9 = KeyCode(char="9")

    # Function keys
    # F13-20 don't work on X11 with pynput because they are not usually
    # part of the keyboard map.
    F1 = virtual_keys.f1
    F2 = virtual_keys.f2
    F3 = virtual_keys.f3
    F4 = virtual_keys.f4
    F5 = virtual_keys.f5
    F6 = virtual_keys.f6
    F7 = virtual_keys.f7
    F8 = virtual_keys.f8
    F9 = virtual_keys.f9
    F10 = virtual_keys.f10
    F11 = virtual_keys.f11
    F12 = virtual_keys.f12
    F13 = virtual_keys.f13
    F14 = virtual_keys.f14
    F15 = virtual_keys.f15
    F16 = virtual_keys.f16
    F17 = virtual_keys.f17
    F18 = virtual_keys.f18
    F19 = virtual_keys.f19
    F20 = virtual_keys.f20


class X11KeySymbols(BasePynputKeySymbols):
    """
    Symbols for X11 from pynput.

    This class defines extra symbols matching those that dragonfly's Win32
    keyboard interface provides.
    """

    # Number pad keys
    # Retrieved from /usr/include/X11/keysymdef.h on Debian 9.
    MULTIPLY = KeyCode.from_vk(0xffaa)
    ADD = KeyCode.from_vk(0xffab)
    SEPARATOR = KeyCode.from_vk(0xffac)
    SUBTRACT = KeyCode.from_vk(0xffad)
    DECIMAL = KeyCode.from_vk(0xffae)
    DIVIDE = KeyCode.from_vk(0xffaf)
    NUMPAD0 = KeyCode.from_vk(0xffb0)
    NUMPAD1 = KeyCode.from_vk(0xffb1)
    NUMPAD2 = KeyCode.from_vk(0xffb2)
    NUMPAD3 = KeyCode.from_vk(0xffb3)
    NUMPAD4 = KeyCode.from_vk(0xffb4)
    NUMPAD5 = KeyCode.from_vk(0xffb5)
    NUMPAD6 = KeyCode.from_vk(0xffb6)
    NUMPAD7 = KeyCode.from_vk(0xffb7)
    NUMPAD8 = KeyCode.from_vk(0xffb8)
    NUMPAD9 = KeyCode.from_vk(0xffb9)

    # Function keys F21-F24.
    # Retrieved from /usr/include/X11/keysymdef.h on Debian 9.
    # These keys don't work on X11 with pynput because they are not usually
    # part of the keyboard map. They are set here to avoid some warnings
    # and because the Windows keyboard supports them.
    F21 = KeyCode.from_vk(0xffd1)
    F22 = KeyCode.from_vk(0xffd2)
    F23 = KeyCode.from_vk(0xffd3)
    F24 = KeyCode.from_vk(0xffd4)

    # Multimedia keys
    # Retrieved from /usr/include/X11/XF86keysym.h on Debian 9.
    # These should work on Debian-based distributions like Ubunutu, but
    # might not work using different X11 server implementations because the
    # symbols are vendor-specific.
    # Any errors raised when typing these or any other keys will be caught
    # and logged.
    VOLUME_UP = KeyCode.from_vk(0x1008FF13)
    VOLUME_DOWN = KeyCode.from_vk(0x1008FF11)
    VOLUME_MUTE = KeyCode.from_vk(0x1008FF12)
    MEDIA_NEXT_TRACK = KeyCode.from_vk(0x1008FF17)
    MEDIA_PREV_TRACK = KeyCode.from_vk(0x1008FF16)
    MEDIA_PLAY_PAUSE = KeyCode.from_vk(0x1008FF14)
    BROWSER_BACK = KeyCode.from_vk(0x1008FF26)
    BROWSER_FORWARD = KeyCode.from_vk(0x1008FF27)


class DarwinKeySymbols(BasePynputKeySymbols):
    """
    Symbols for Darwin from pynput.
    """

    # The key symbols below are defined solely to prevent errors in
    #  typeables.py.  ValueErrors will be raised by send_keyboard_events()
    #  if these symbols are given in an event sequence.

    # Extra function keys.
    F21 = virtual_keys.f21
    F22 = virtual_keys.f22
    F23 = virtual_keys.f23
    F24 = virtual_keys.f24

    # Multimedia keys.
    VOLUME_UP = virtual_keys.volume_up
    VOLUME_DOWN = virtual_keys.volume_down
    VOLUME_MUTE = virtual_keys.volume_mute
    MEDIA_NEXT_TRACK = virtual_keys.media_next_track
    MEDIA_PREV_TRACK = virtual_keys.media_prev_track
    MEDIA_PLAY_PAUSE = virtual_keys.media_play_pause
    BROWSER_BACK = virtual_keys.browser_back
    BROWSER_FORWARD = virtual_keys.browser_forward


class PynputKeyboard(BaseKeyboard):
    """Static class wrapper around pynput.keyboard."""

    _controller = Controller()

    @classmethod
    def send_keyboard_events(cls, events):
        """
        Send a sequence of keyboard events.

        Positional arguments:
        events -- a sequence of tuples of the form
            (keycode, down, timeout), where
                keycode (str|KeyCode): pynput key code.
                down (boolean): True means the key will be pressed down,
                    False means the key will be released.
                timeout (int): number of seconds to sleep after
                    the keyboard event.

        """
        cls._log.debug("Keyboard.send_keyboard_events %r", events)
        for event in events:
            (key, down, timeout) = event

            # Raise an error if the key is unsupported. 'key' can also be a
            # string, e.g. "a", "b", "/", etc, but we don't check if those
            # are valid.
            if isinstance(key, KeyCode) and key.vk == -1:
                raise ValueError("Unsupported key: %r" % key.char)

            # Press/release the key, catching any errors.
            try:
                cls._controller.touch(key, down)
            except Exception as e:
                cls._log.exception("Failed to type key code %s: %s",
                                   key, e)

            # Sleep after the keyboard event if necessary.
            if timeout:
                time.sleep(timeout)

    @classmethod
    def get_typeable(cls, char, is_text=False):
        return PynputTypeable(char, is_text=is_text)



from contextlib import contextmanager
import sys

from . import controller

from .utils import (CursorPosition, TextQuery)

if sys.platform.startswith("win"):
    from . import ia2
    os_controller_class = ia2.Controller
elif sys.platform.startswith("linux"):
    from . import atspi
    os_controller_class = atspi.Controller
else:
    os_controller_class = None

controller_instance = None

def get_accessibility_controller():
    """Get the OS-independent accessibility controller which is the gateway to all
    accessibility functionality. Returns None if OS is not supported."""

    global controller_instance
    if os_controller_class and (not controller_instance or controller_instance.stopped):
        os_controller = os_controller_class()
        controller_instance = controller.AccessibilityController(os_controller)
    return controller_instance

@contextmanager
def get_stopping_accessibility_controller():
    """Same as :func:`get_accessibility_controller`, but automatically stops when
    used in a `with` context."""

    yield get_accessibility_controller()
    if controller_instance:
        controller_instance.stop()



# encoding: utf-8
#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

# pylint: disable=line-too-long

"""
Text action
============================================================================

This section describes the :class:`Text` action object. This type of
action is used for typing text into the foreground application.  This works
on Windows, Mac OS and with X11 (e.g. on Linux).

To use this class on X11/Linux, the
`xdotool <https://www.semicomplete.com/projects/xdotool/>`__ program must be
installed and the ``XDG_SESSION_TYPE`` environment variable set to ``x11``.
This class does **not** support typing text in Wayland sessions.

It differs from the :class:`Key` action in that :class:`Text` is used for
typing literal text, while :class:`dragonfly.actions.action_key.Key`
emulates pressing keys on the keyboard.  An example of this is that the
arrow-keys are not part of a text and so cannot be typed using the
:class:`Text` action, but can be sent by the
:class:`dragonfly.actions.action_key.Key` action.


Windows Unicode Keyboard Support
............................................................................

The :class:`Text` action can be used to type arbitrary Unicode characters
using the `relevant Windows API <https://docs.microsoft.com/en-us/windows/desktop/api/winuser/ns-winuser-tagkeybdinput#remarks>`__.
This is disabled by default because it ignores the up/down status of
modifier keys (e.g. ctrl).

It can be enabled by changing the ``unicode_keyboard`` setting in
`~/.dragonfly2-speech/settings.cfg` to ``True``::

    unicode_keyboard = True

If you need to simulate typing arbitrary Unicode characters *and* have
*individual* :class:`Text` actions respect modifier keys normally for normal
characters, set the configuration as above and use the ``use_hardware``
parameter for :class:`Text` as follows:

.. code:: python

   action = Text("σμ") + Key("ctrl:down") + Text("]", use_hardware=True) + Key("ctrl:up")
   action.execute()

Some applications require hardware emulation versus Unicode keyboard
emulation. If you use such applications, add their executable names to the
``hardware_apps`` list in the configuration file mentioned above to make
dragonfly always use hardware emulation for them.

If hardware emulation is required, then the action will use the keyboard
layout of the foreground window when calculating keyboard events. If any of
the specified characters are not typeable using the current window's
keyboard layout, then an error will be logged and no keys will be typed::

    action.exec (ERROR): Execution failed: Keyboard interface cannot type this character: 'μ'

Keys in ranges 0-9, a-z and A-Z are always typeable. If keys in these ranges
cannot be typed using the current keyboard layout, then the equivalent key
will be used instead. For example, the following code will result in the 'я'
key being pressed when using the main Cyrillic keyboard layout: ::

   # This is equivalent to Text(u"яЯ").
   Text("zZ").execute()

These settings and parameters have no effect on other platforms.


X11/Linux Unicode Keyboard Support
............................................................................

The :class:`Text` action can also type arbitrary Unicode characters on X11.
This works regardless of the ``use_hardware`` parameter or
``unicode_keyboard`` setting.

Unlike on Windows, modifier keys will be respected by :class:`Text` actions
on X11. As such, the previous Windows example will work and can even be
simplified a little:

.. code:: python

   action = Text("σμ") + Key("ctrl:down") + Text("]") + Key("ctrl:up")
   action.execute()


It can also be done with one :class:`Key` action:

.. code:: python

   Key("σ,μ,c-]").execute()


Text class reference
............................................................................

"""

from locale import getpreferredencoding

from six import binary_type

from ..engines import get_engine
from ..util.clipboard import Clipboard
from .action_base import ActionError
from .action_base_keyboard import BaseKeyboardAction
from .action_key import Key

#---------------------------------------------------------------------------


class Text(BaseKeyboardAction):
    """
    `Action` that sends keyboard events to type text.

    Arguments:
     - *spec* (*str*) -- the text to type
     - *static* (boolean) --
       if *True*, do not dynamically interpret *spec*
       when executing this action
     - *pause* (*float*) --
       the time to pause between each keystroke, given
       in seconds
     - *autofmt* (boolean) --
       if *True*, attempt to format the text with correct
       spacing and capitalization.  This is done by first mimicking
       a word recognition and then analyzing its spacing and
       capitalization and applying the same formatting to the text.
     - *use_hardware* (boolean) --
       if *True*, send keyboard events using hardware emulation instead of
       as Unicode text. This will respect the up/down status of modifier
       keys.
    """

    _specials = {
        "\n": "enter",
        "\t": "tab",
    }

    def __init__(self, spec=None, static=False, pause=None,
                 autofmt=False, use_hardware=False):
        if isinstance(spec, binary_type):
            spec = spec.decode(getpreferredencoding())
        BaseKeyboardAction.__init__(self, spec=spec, static=static,
                                    use_hardware=use_hardware)
        # Set other members.
        self._autofmt = autofmt
        self._pause = self._pause_default if pause is None else pause

    def _parse_spec(self, spec):
        """Convert the given *spec* to keyboard events."""
        key_symbols = []
        for character in spec:
            # The character is the key symbol, except in special cases.
            if character in self._specials:
                key_symbol = self._specials[character]
            else:
                key_symbol = character

            # Add the key symbol to the list.
            key_symbols.append(key_symbol)

        return key_symbols

    def _execute_events(self, events):
        """
        Send keyboard events.

        If instance was initialized with *autofmt* True,
        then this method will mimic a word recognition
        and analyze its formatting so as to autoformat
        the text's spacing and capitalization before
        sending it as keyboard events.
        """
        if self._autofmt:
            # Mimic a word, select and copy it to retrieve capitalization.
            get_engine().mimic("test")
            Key("cs-left, c-c/5").execute()
            word = Clipboard.get_system_text()

            # Inspect formatting of the mimicked word.
            index = word.find("test")
            if index == -1:
                index = word.find("Test")
                capitalize = True
                if index == -1:
                    self._log.error("Failed to autoformat.")
                    return False
            else:
                capitalize = False

            # Capitalize given text if necessary.
            text = self._spec
            if capitalize:
                text = text[0].capitalize() + text[1:]

            # Reconstruct autoformatted output and convert it
            #  to keyboard events.
            prefix = word[:index]
            suffix = word[index + 4:]
            events = self._parse_spec(prefix + text + suffix)

        # Calculate keyboard events.
        use_hardware = self.require_hardware_events()
        keyboard_events = []
        for key_symbol in events:
            # Get a Typeable object for each key symbol, if possible.
            typeable = self._get_typeable(key_symbol, use_hardware)

            # Raise an error message if a Typeable could not be retrieved.
            if typeable is None:
                error_message = ("Keyboard interface cannot type this"
                                 " character: %r" % key_symbol)
                raise ActionError(error_message)

            # Get keyboard events using the Typeable.
            keyboard_events.extend(typeable.events(self._pause))

        # Send keyboard events.
        self._keyboard.send_keyboard_events(keyboard_events)
        return True

    def __str__(self):
        return u"{!r}".format(self._spec)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

# The overridden methods in the Dragonfly.list.List class were
# automatically generated using the commented-out code immediately below.
#
#def construct_skeleton():
#   instance = list()
#   ignored_attributes = ('__class__', '__contains__', '__doc__',
#       '__eq__', '__ge__', '__getattribute__', '__getitem__',
#       '__getslice__', '__gt__', '__hash__', '__init__', '__iter__',
#       '__le__', '__len__', '__lt__', '__ne__', '__new__', '__repr__',
#       '__str__', 'count', 'index', '__delattr__', '__setattr__')
#   output = []
#   for name in dir(instance):
#       attribute = getattr(instance, name)
#       if not callable(attribute): continue
#       if name in ignored_attributes: continue
#       output.append("""\
#   def %(function)s(self, *args, **kwargs):
#       result = %(class)s.%(function)s(self, *args, **kwargs)
#       self._update(); return result
#""" % {"class": "list", "function": name})
#   return "".join(output)
#print construct_skeleton()
from six import string_types

#===========================================================================
# Base class for dragonfly list objects.

class ListBase(object):
    """ Base class for dragonfly list objects. """

    def __init__(self, name):
        self._name = name
        self._grammar = None
        self._batch_mode = False
        self._batch_updates = False

    #-----------------------------------------------------------------------
    # Protected attribute access.

    valid_types = property(
        lambda self: string_types,
        doc="The types of object at a Dragonfly list can contain.")

    name = property(lambda self: self._name,
                    doc="Read-only access to a list's name.")

    def _get_grammar(self):
        return self._grammar

    def _set_grammar(self, grammar):
        self._grammar = grammar
        # if self._grammar is None:
        #     self._grammar = grammar
        # else:
        #     raise TypeError("The grammar object a Dragonfly list is bound "
        #                     "to cannot be changed after it has been set.")

    grammar = property(_get_grammar, _set_grammar,
                       doc="Set-once access to a list's grammar object.")

    #-----------------------------------------------------------------------
    # Context manager methods for optimizing update_list() calls.

    def __enter__(self):
        self._batch_mode = True

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._batch_mode = False
        if self._batch_updates:
            self._update()
            self._batch_updates = False

    #-----------------------------------------------------------------------
    # Notify the grammar of a list modification.

    def _update(self):
        """
        Internal method that notifies the engine of list updates.

        This method should be called internally by :class:`ListBase`sub-
        classes when the list is modified.
        """
        # Return early for batch mode. A single update_list() call will
        # occur in __exit__(), after a 'with' block.
        if self._batch_mode:
            self._batch_updates = True
            return

        # Validate list items.
        self._validate_items()

        # If this list is part of a grammar, then notify it of the list
        # changes.
        if self._grammar:
            self._grammar.update_list(self)

    def _validate_items(self):
        valid_types = self.valid_types
        invalid = [i for i in self if not isinstance(i, valid_types)]
        if invalid:
            raise TypeError("Dragonfly lists can only contain"
                            " string objects; received: %r" % invalid)

    #-----------------------------------------------------------------------
    # Accessor for the grammar to retrieve the list items.
    def get_list_items(self):
        raise NotImplementedError("Call to virtual method list_items()")


#===========================================================================
# Wrapper for Python's built-in list type.

class List(ListBase, list):
    """
        Wrapper for Python's built-in list that supports automatic
        engine notification of changes.

        Use :class:`~dragonfly.grammar.elements_basic.ListRef` elements
        in a grammar rule to allow matching speech to list items.
    """

    def __init__(self, name, *args, **kwargs):
        ListBase.__init__(self, name)
        list.__init__(self, *args, **kwargs)

    #-----------------------------------------------------------------------
    # Accessor for the grammar to retrieve the list items.

    def get_list_items(self):
        return self

    #-----------------------------------------------------------------------
    # Custom methods.

    def set(self, other):
        """Set the contents of this list to the contents of another."""
        with self:
            self[:] = other

    #-----------------------------------------------------------------------
    # Overridden list methods.

    def __add__(self, *args, **kwargs):
        result = list.__add__(self, *args, **kwargs)
        self._update(); return result
    def __delitem__(self, *args, **kwargs):
        result = list.__delitem__(self, *args, **kwargs)
        self._update(); return result
    def __delslice__(self, *args, **kwargs):
        # pylint: disable=no-member
        result = list.__delslice__(self, *args, **kwargs)
        self._update(); return result
    def __iadd__(self, *args, **kwargs):
        result = list.__iadd__(self, *args, **kwargs)
        self._update(); return result
    def __imul__(self, *args, **kwargs):
        result = list.__imul__(self, *args, **kwargs)
        self._update(); return result
    def __mul__(self, *args, **kwargs):
        result = list.__mul__(self, *args, **kwargs)
        self._update(); return result
    def __reduce__(self, *args, **kwargs):
        result = list.__reduce__(self, *args, **kwargs)
        self._update(); return result
    def __reduce_ex__(self, *args, **kwargs):
        result = list.__reduce_ex__(self, *args, **kwargs)
        self._update(); return result
    def __rmul__(self, *args, **kwargs):
        result = list.__rmul__(self, *args, **kwargs)
        self._update(); return result
    def __setitem__(self, *args, **kwargs):
        result = list.__setitem__(self, *args, **kwargs)
        self._update(); return result
    def __setslice__(self, *args, **kwargs):
        # pylint: disable=no-member
        result = list.__setslice__(self, *args, **kwargs)
        self._update(); return result
    def append(self, *args, **kwargs):
        result = list.append(self, *args, **kwargs)
        self._update(); return result
    def extend(self, *args, **kwargs):
        result = list.extend(self, *args, **kwargs)
        self._update(); return result
    def insert(self, *args, **kwargs):
        result = list.insert(self, *args, **kwargs)
        self._update(); return result
    def pop(self, *args, **kwargs):
        result = list.pop(self, *args, **kwargs)
        self._update(); return result
    def remove(self, *args, **kwargs):
        result = list.remove(self, *args, **kwargs)
        self._update(); return result
    def reverse(self, *args, **kwargs):
        result = list.reverse(self, *args, **kwargs)
        self._update(); return result
    def sort(self, *args, **kwargs):
        result = list.sort(self, *args, **kwargs)
        self._update(); return result
    def clear(self):
        del self[:]

#===========================================================================
# Wrapper for Python's built-in dict type.

class DictList(ListBase, dict):
    """
        Wrapper for Python's built-in dict that supports automatic
        engine notification of changes.  The object's keys are used
        as the elements of the engine list, while use of the associated
        values is left to the user.

        Use :class:`~dragonfly.grammar.elements_basic.DictListRef` elements
        in a grammar rule to allow matching speech to dictionary keys.
    """

    def __init__(self, name, *args, **kwargs):
        ListBase.__init__(self, name)
        dict.__init__(self, *args, **kwargs)

    #-----------------------------------------------------------------------
    # Accessor for the grammar to retrieve the list items.

    def get_list_items(self):
        return list(self.keys())

    #-----------------------------------------------------------------------
    # Custom methods.

    def set(self, other):
        """Set the contents of this dict to the contents of another."""
        with self:
            self.clear()
            self.update(other)

    #-----------------------------------------------------------------------
    # Overridden dict methods.

    def __delitem__(self, *args, **kwargs):
        result = dict.__delitem__(self, *args, **kwargs)
        self._update(); return result
    def __reduce__(self, *args, **kwargs):
        result = dict.__reduce__(self, *args, **kwargs)
        self._update(); return result
    def __reduce_ex__(self, *args, **kwargs):
        result = dict.__reduce_ex__(self, *args, **kwargs)
        self._update(); return result
    def __setitem__(self, *args, **kwargs):
        result = dict.__setitem__(self, *args, **kwargs)
        self._update(); return result
    def clear(self, *args, **kwargs):
        result = dict.clear(self, *args, **kwargs)
        self._update(); return result
    def fromkeys(self, *args, **kwargs):
        result = dict.fromkeys(self, *args, **kwargs)
        self._update(); return result
    def pop(self, *args, **kwargs):
        result = dict.pop(self, *args, **kwargs)
        self._update(); return result
    def popitem(self, *args, **kwargs):
        result = dict.popitem(self, *args, **kwargs)
        self._update(); return result
    def setdefault(self, *args, **kwargs):
        result = dict.setdefault(self, *args, **kwargs)
        self._update(); return result
    def update(self, *args, **kwargs):
        result = dict.update(self, *args, **kwargs)
        self._update(); return result



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Engine class for CMU Pocket Sphinx
"""

import contextlib
import locale
import os
import wave

from six import binary_type, text_type, string_types, PY2
from jsgf import RootGrammar, PublicRule, Literal
from sphinxwrapper import PocketSphinx

from dragonfly import Window
from ..base import (EngineBase, EngineError, MimicFailure,
                    DelegateTimerManagerInterface,
                    DictationContainerBase)
from .compiler import SphinxJSGFCompiler
from .grammar_wrapper import GrammarWrapper
from .misc import (EngineConfig, WaveRecognitionObserver,
                   get_decoder_config_object)
from .recobs import SphinxRecObsManager
from .recording import PyAudioRecorder
from .timer import SphinxTimerManager
from .training import write_training_data, write_transcript_files


class UnknownWordError(Exception):
    pass


def _map_to_str(text, encoding=locale.getpreferredencoding()):
    # Translate unicode/bytes to whatever str is in this version of
    # Python. Decoder methods seem to require str objects.
    if not isinstance(text, string_types):
        text = str(text)
    if PY2 and isinstance(text, text_type):
        text = text.encode(encoding)
    elif not PY2 and isinstance(text, binary_type):
        text = text.decode(encoding)
    return text


class SphinxEngine(EngineBase, DelegateTimerManagerInterface):
    """ Speech recognition engine back-end for CMU Pocket Sphinx. """

    _name = "sphinx"
    DictationContainer = DictationContainerBase

    def __init__(self):
        EngineBase.__init__(self)
        DelegateTimerManagerInterface.__init__(self)

        try:
            import sphinxwrapper, jsgf, pyaudio
        except ImportError as e:
            raise EngineError("Failed to import Pocket Sphinx engine "
                              "dependencies.")

        # Set the default engine configuration.
        # This can be changed later using the config property.
        self._config = None
        self.config = EngineConfig

        # Set other variables
        self._decoder = None
        self._audio_buffers = []
        self.compiler = SphinxJSGFCompiler(self)
        self._recognition_observer_manager = SphinxRecObsManager(self)
        self._keyphrase_thresholds = {}
        self._keyphrase_functions = {}
        self._training_session_active = False
        self._default_search_result = None
        self._grammar_count = 0

        # Timer-related members.
        self._timer_manager = SphinxTimerManager(0.02, self)

        # Set up keyphrase search names and valid search names for grammars.
        self._keyphrase_search_names = ["_key_phrases", "_wake_phrase"]
        self._valid_searches = set()

        # Recognising loop members.
        self._recorder = PyAudioRecorder(self.config)
        self._cancel_recognition_next_time = False
        self._recognising = False
        self._recognition_paused = False

    @property
    def config(self):
        """
        Python module/object containing engine configuration.

        You will need to restart the engine with :meth:`disconnect` and
        :meth:`connect` if the configuration has been changed after
        :meth:`connect` has been called.

        :returns: config module/object
        """
        return self._config

    @config.setter
    def config(self, value):
        # Validate configuration object.
        self.validate_config(value)
        self._config = value

    @classmethod
    def validate_config(cls, engine_config):
        # Check configuration options and set defaults where appropriate.
        # Set a new decoder config if necessary.
        if not hasattr(engine_config, "DECODER_CONFIG"):
            setattr(engine_config, "DECODER_CONFIG",
                    get_decoder_config_object())
        options = [
            "LANGUAGE",

            "START_ASLEEP",
            "WAKE_PHRASE",
            "WAKE_PHRASE_THRESHOLD",
            "SLEEP_PHRASE",
            "SLEEP_PHRASE_THRESHOLD",

            "TRAINING_DATA_DIR",
            "TRANSCRIPT_NAME",
            "START_TRAINING_PHRASE",
            "START_TRAINING_PHRASE_THRESHOLD",
            "END_TRAINING_PHRASE",
            "END_TRAINING_PHRASE_THRESHOLD",

            "CHANNELS",
            "RATE",
            "SAMPLE_WIDTH",
            "FRAMES_PER_BUFFER",
        ]

        # Get default values and set them they are missing.
        for option in options:
            if hasattr(engine_config, option):
                continue

            default_value = getattr(EngineConfig, option)
            if "PHRASE" in option:
                # Disable missing phrases by default if using a language
                # other than English.
                if not engine_config.LANGUAGE.startswith("en"):
                    default_value = "" if option.endswith("PHRASE") else 0.0
            setattr(engine_config, option, default_value)

    def connect(self):
        """
        Set up the CMU Pocket Sphinx decoder.

        This method does nothing if the engine is already connected.
        """
        if self._decoder:
            return

        # Initialise a new decoder with the given configuration
        decoder_config = self._config.DECODER_CONFIG
        self._decoder = PocketSphinx(decoder_config)
        self._valid_searches.add(self._default_search_name)

        # Set up callback function wrappers
        def hypothesis(hyp):
            # Set default search result.
            self._default_search_result = hyp

            # Set speech to the hypothesis string or None if there isn't one
            speech = hyp.hypstr if hyp else None
            return self._hypothesis_callback(speech, False)

        def speech_start():
            # Reset the default search result and call the engine's callback
            # method.
            self._default_search_result = None
            return self._speech_start_callback(False)

        self._decoder.hypothesis_callback = hypothesis
        self._decoder.speech_start_callback = speech_start

        # Set up built-in keyphrases if they set. Catch and log any
        # UnknownWordErrors because all keyphrases are optional.
        def get_phrase_values(name):
            phrase_attr = name + "_PHRASE"
            threshold_attr = name + "_PHRASE_THRESHOLD"
            return (getattr(self.config, phrase_attr, ""),
                    getattr(self.config, threshold_attr, 0))

        def safe_set_keyphrase(name, func):
            phrase, threshold = get_phrase_values(name)
            if phrase and threshold:
                try:
                    self.set_keyphrase(phrase, threshold, func)
                except UnknownWordError as e:
                    self._log.error(e)

        # Set the wake phrase using set_kws_list directly because it uses a
        # different search.
        wake_phrase, wake_threshold = get_phrase_values("WAKE")
        if wake_phrase and wake_threshold:
            try:
                self._validate_words(wake_phrase.split(), "keyphrase")
                self._decoder.set_kws_list("_wake_phrase", {
                    wake_phrase: wake_threshold
                })
            except UnknownWordError as e:
                self._log.error(e)

        # Set the other keyphrases using safe_set_keyphrase().
        safe_set_keyphrase("SLEEP", self.pause_recognition)
        safe_set_keyphrase("START_TRAINING",
                           self.start_training_session)
        safe_set_keyphrase("END_TRAINING",
                           self.end_training_session)

        # Set the PyAudioRecorder instance's config object.
        self._recorder.config = self.config

        # Start in sleep mode if requested.
        if self.config.START_ASLEEP:
            self.pause_recognition()
            self._log.warning("Starting in sleep mode as requested.")

    def _free_engine_resources(self):
        """
        Internal method for freeing the resources used by the engine.
        """
        # Stop the audio recorder if it is running.
        self._recognising = False
        self._recorder.stop()

        # Free the decoder and clear audio buffers.
        self._decoder = None
        self._audio_buffers = []

        # Reset other variables
        self._cancel_recognition_next_time = False
        self._training_session_active = False
        self._recognition_paused = False
        self._grammar_count = 0

        # Clear dictionaries and sets
        self._grammar_wrappers.clear()
        self._valid_searches.clear()
        self._keyphrase_thresholds.clear()
        self._keyphrase_functions.clear()

    def disconnect(self):
        """
        Deallocate the CMU Sphinx decoder and any other resources used by
        it.

        This method effectively unloads all loaded grammars and key
        phrases.
        """
        # Free resources if the decoder isn't currently being used to
        # recognise, otherwise stop the recognising loop, which will free
        # the resources safely.
        if not self.recognising:
            self._free_engine_resources()
        else:
            self._recognising = False
            self._recorder.stop()

    # -----------------------------------------------------------------------
    # Multiplexing timer methods.

    def create_timer(self, callback, interval, repeating=True):
        """
        Create and return a timer using the specified callback and repeat
        interval.

        **Note**: Timers will not run unless the engine is recognising
        audio. Normal threads can be used instead with no downsides.
        """
        if not self.recognising:
            self._log.warning("Timers will not run unless the engine is "
                              "recognising audio.")

        return super(SphinxEngine, self).create_timer(callback, interval,
                                                      repeating)

    # -----------------------------------------------------------------------
    # Methods for working with grammars.

    def check_valid_word(self, word):
        """
        Check if a word is in the current Sphinx pronunciation dictionary.

        :rtype: bool
        """
        if not self._decoder:
            self.connect()

        word = _map_to_str(word)
        return bool(self._decoder.lookup_word(word.lower()))

    def _validate_words(self, words, search_type):
        unknown_words = []

        # Use 'set' to de-duplicate the 'words' list.
        for word in set(words):
            if not self.check_valid_word(word):
                unknown_words.append(word)

        if unknown_words:
            # Sort the word list before using it.
            unknown_words.sort()
            raise UnknownWordError(
                "%s used words not found in the pronunciation dictionary: "
                "%s" % (search_type, ", ".join(unknown_words))
            )

    def _build_grammar_wrapper(self, grammar):
        search_name = "%d" % self._grammar_count
        self._grammar_count += 1
        return GrammarWrapper(grammar, self,
                              self._recognition_observer_manager,
                              search_name)

    def _set_grammar(self, wrapper, activate, partial=False):
        if not wrapper:
            return

        # Connect to the engine if it isn't connected already.
        self.connect()

        def activate_search_if_necessary():
            if activate:
                self._decoder.end_utterance()
                self._decoder.active_search = wrapper.search_name

        # Check if the wrapper's search name is valid.
        # Set the search (again) if necessary.
        valid_search = wrapper.search_name in self._valid_searches
        if valid_search and not wrapper.set_search:
            # wrapper.search_name is a valid search, so return.
            activate_search_if_necessary()
            return

        # Return early if 'partial' is True as an optimisation to avoid
        # recompiling grammars for every rule activation/deactivation.
        # Also return if the search doesn't need to be set.
        if partial or not wrapper.set_search:
            return

        # Compile and set the jsgf search.
        compiled = wrapper.compile_jsgf()
        self._log.debug(compiled)

        # Raise an error if there are no active public rules.
        if "public <root> = " not in compiled:
            raise EngineError("no public rules found in the grammar")

        # Set the JSGF search.
        self._decoder.end_utterance()
        self._decoder.set_jsgf_string(wrapper.search_name,
                                      _map_to_str(compiled))
        activate_search_if_necessary()

        # Grammar search has been loaded, so set the wrapper's flag.
        wrapper.set_search = False

    def _unset_search(self, name):
        # Unset a Pocket Sphinx search with the given name.
        # Don't unset the default or keyphrase searches.
        default_search = self._default_search_name
        reserved = [default_search] + self._keyphrase_search_names
        if name in reserved:
            return

        # Unset the Pocket Sphinx search.
        if name in self._valid_searches:
            # Unset the decoder search.
            self._decoder.unset_search(name)

            # Remove the search from the valid searches set.
            self._valid_searches.remove(name)

        # Change to the default search to avoid possible segmentation faults
        # from Pocket Sphinx which crash Python.
        self._set_default_search()

    # TODO Add optional context parameter
    def set_keyphrase(self, keyphrase, threshold, func):
        """
        Add a keyphrase to listen for.

        Key phrases take precedence over grammars as they are processed first.
        They cannot be set for specific contexts (yet).

        :param keyphrase: keyphrase to add.
        :param threshold: keyphrase threshold value to use.
        :param func: function or method to call when the keyphrase is heard.
        :type keyphrase: str
        :type threshold: float
        :type func: callable
        :raises: UnknownWordError
        """
        # Check that all words in the keyphrase are in the pronunciation dictionary.
        # This can raise an UnknownWordError.
        self._validate_words(keyphrase.split(), "keyphrase")

        # Check that the threshold is a float.
        if not isinstance(threshold, float):
            raise TypeError("threshold must be a float, not %s" % threshold)

        # Add parameters to the relevant dictionaries.
        self._keyphrase_thresholds[keyphrase] = threshold
        self._keyphrase_functions[keyphrase] = func

        # Set the keyphrase search (again)
        self._decoder.end_utterance()
        self._decoder.set_kws_list("_key_phrases", self._keyphrase_thresholds)

    def unset_keyphrase(self, keyphrase):
        """
        Remove a set keyphrase so that the engine no longer listens for it.

        :param keyphrase: keyphrase to remove.
        :type keyphrase: str
        """
        # Remove parameters from the relevant dictionaries. Don't raise an error
        # if there is no such keyphrase.
        self._keyphrase_thresholds.pop(keyphrase, None)
        self._keyphrase_functions.pop(keyphrase, None)

        # Set the keyphrase search (again)
        self._decoder.end_utterance()
        self._decoder.set_kws_list("_key_phrases", self._keyphrase_thresholds)

    def _set_default_search(self):
        # Change the active search to the one used for processing speech as
        # it is heard.
        swap_to_wake_search = (
            self.recognition_paused and self.config.WAKE_PHRASE and
            self.config.WAKE_PHRASE_THRESHOLD
        )

        # Ensure we're not processing.
        self._decoder.end_utterance()
        if swap_to_wake_search:
            self._decoder.active_search = "_wake_phrase"
        else:
            self._decoder.active_search = self._default_search_name

    def _load_grammar(self, grammar):
        """ Load the given *grammar* and return a wrapper. """
        self._log.debug("Engine %s: loading grammar %s."
                        % (self, grammar.name))
        wrapper = self._build_grammar_wrapper(grammar)

        # Attempt to set the grammar search.
        try:
            self._set_grammar(wrapper, False)
        except Exception as e:
            self._log.exception("Failed to load grammar %s: %s."
                                % (grammar, e))
            raise EngineError("Failed to load grammar %s: %s."
                              % (grammar, e))

        # Set the grammar wrapper's search name as valid and return the
        # wrapper.
        self._valid_searches.add(wrapper.search_name)
        return wrapper

    def _unload_grammar(self, grammar, wrapper):
        try:
            # Unset the search names for the grammar.
            self._unset_search(wrapper.search_name)
        except Exception as e:
            self._log.exception("Failed to unload grammar %s: %s."
                                % (grammar, e))

    def activate_grammar(self, grammar):
        self._log.debug("Activating grammar %s." % grammar.name)

    def deactivate_grammar(self, grammar):
        self._log.debug("Deactivating grammar %s." % grammar.name)

    def activate_rule(self, rule, grammar):
        self._log.debug("Activating rule %s in grammar %s."
                        % (rule.name, grammar.name))
        wrapper = self._get_grammar_wrapper(grammar)
        if not wrapper:
            return
        try:
            wrapper.enable_rule(rule.name)
            self._set_grammar(wrapper, False, True)
        except Exception as e:
            self._log.exception("Failed to activate grammar %s: %s."
                                % (grammar, e))

    def deactivate_rule(self, rule, grammar):
        self._log.debug("Deactivating rule %s in grammar %s."
                        % (rule.name, grammar.name))
        wrapper = self._get_grammar_wrapper(grammar)
        if not wrapper:
            return
        try:
            wrapper.disable_rule(rule.name)
            self._set_grammar(wrapper, False, True)
        except Exception as e:
            self._log.exception("Failed to activate grammar %s: %s."
                                % (grammar, e))

    def update_list(self, lst, grammar):
        wrapper = self._get_grammar_wrapper(grammar)
        if not wrapper:
            return

        # Unfortunately there is no way to update lists for Pocket Sphinx
        # without reloading the grammar, so we'll update the list's JSGF
        # rule and reload.
        wrapper.update_list(lst)

        # Reload the grammar.
        try:
            self._set_grammar(wrapper, False)
        except Exception as e:
            self._log.exception("Failed to update list %s: %s."
                                % (lst, e))

    def set_exclusiveness(self, grammar, exclusive):
        wrapper = self._get_grammar_wrapper(grammar)
        if not wrapper:
            return

        wrapper.exclusive = exclusive

    # -----------------------------------------------------------------------
    # Miscellaneous methods.

    @property
    def recognising(self):
        """
        Whether the engine is currently recognising speech.

        To stop recognition, use :meth:`disconnect`.

        :rtype: bool
        """
        return self._recorder.recording or self._recognising

    @property
    def default_search_result(self):
        """
        The last hypothesis object of the default search.

        This does not currently reach recognition observers because it is
        intended to be used for dictation results, which are currently
        disabled. Nevertheless this object can be useful sometimes.

        :returns: Sphinx Hypothesis object | None
        """
        return self._default_search_result

    @property
    def _default_search_name(self):
        # The name of the Pocket Sphinx search used for processing speech as
        # it is heard.
        return "_default"

    def _get_best_hypothesis(self, hypotheses):
        """
        Take a list of speech hypotheses and return the most likely one.

        :type hypotheses: iterable
        :return: str | None
        """
        # Get all distinct, non-null hypotheses.
        distinct = tuple([h for h in set(hypotheses) if bool(h)])
        if not distinct:
            return None
        elif len(distinct) == 1:
            return distinct[0]  # only one choice

        # Decide between non-null hypotheses using a Pocket Sphinx search with
        # each hypothesis as a grammar rule.
        grammar = RootGrammar()
        grammar.language_name = self.language
        for i, hypothesis in enumerate(distinct):
            grammar.add_rule(PublicRule("rule%d" % i, Literal(hypothesis)))

        compiled = grammar.compile_grammar()
        name = "_temp"

        # Store the current search name.
        original = self._decoder.active_search

        # Note that there is no need to validate words in this case because
        # each literal in the _temp grammar came from a Pocket Sphinx
        # hypothesis.
        self._decoder.end_utterance()
        self._decoder.set_jsgf_string(name, _map_to_str(compiled))
        self._decoder.active_search = name

        # Do the processing.
        hyp = self._decoder.batch_process(
            self._audio_buffers,
            use_callbacks=False
        )
        result = hyp.hypstr if hyp else None

        # Switch back to the previous search.
        self._decoder.end_utterance()  # just in case
        self._decoder.active_search = original
        self._decoder.unset_search("_temp")
        return result

    def _speech_start_callback(self, mimicking):
        # Get context info.
        fg_window = Window.get_foreground()
        window_info = {
            "executable": fg_window.executable,
            "title": fg_window.title,
            "handle": fg_window.handle,
        }

        # Call process_begin for all grammars so that any out of context
        # grammar will not be used.
        for wrapper in self._grammar_wrappers.copy().values():
            wrapper.process_begin(**window_info)

        if not mimicking:
            # Trim excess audio buffers from the start of the list. Keep a maximum 1
            # second of silence before speech start was detected. This should help
            # increase the performance of batch reprocessing later.
            chunk = self.config.FRAMES_PER_BUFFER
            rate = self.config.RATE
            seconds = 1
            n_buffers = int(rate / chunk * seconds)
            self._audio_buffers = self._audio_buffers[-1 * n_buffers:]

        # Notify observers
        self._recognition_observer_manager.notify_begin()

    def _hypothesis_callback(self, speech, mimicking):
        """
        Internal Pocket Sphinx hypothesis callback method. Calls _process_hypothesis
        and does post-processing afterwards.
        :param speech: speech hypothesis
        :type speech: str | None
        :param mimicking:  whether to treat speech as mimicked speech.
        :rtype: bool
        """
        # Clear any recorded audio buffers.
        self._recorder.clear_buffers()

        # Process speech. We should get back a boolean for whether processing
        # occurred as well as the final speech hypothesis.
        processing_occurred, final_speech = self._process_hypotheses(
            speech, mimicking
        )

        # Notify observers of failure.
        results_obj = None  # TODO Use PS results object once implemented
        if not processing_occurred:
            self._recognition_observer_manager.notify_failure(results_obj)

        # Write the training data files if necessary.
        data_dir = self.config.TRAINING_DATA_DIR
        if not mimicking and data_dir and os.path.isdir(data_dir):
            # Use the default search's hypothesis if final_speech was nil.
            if not final_speech:
                final_speech = speech
            try:
                write_training_data(self.config, self._audio_buffers,
                                    final_speech)
            except Exception as e:
                self._log.exception("Failed to write training data: %s" % e)

        # Clear audio buffer list because utterance processing has finished.
        self._audio_buffers = []

        # Ensure that the correct search is used.
        self._set_default_search()

        # Return whether processing occurred in case this method was called
        # by mimic.
        return processing_occurred

    def _process_key_phrases(self, speech, mimicking):
        """
        Processing key phrase searches and return the matched keyphrase
        (if any).

        :type speech: str
        :param mimicking: whether to treat speech as mimicked speech.
        :rtype: str
        """
        # Return if speech is empty/null or if there are no key phrases set.
        if not (speech and self._keyphrase_thresholds):
            return ""  # no matches

        if not mimicking:
            # Reprocess using the key phrases search
            self._decoder.end_utterance()
            self._decoder.active_search = "_key_phrases"
            hyp = self._decoder.batch_process(self._audio_buffers,
                                              use_callbacks=False)

            # Get the hypothesis string.
            speech = hyp.hypstr if hyp else ""

            # Restore search to the default search.
            self._set_default_search()

            # Return if no key phrase matched.
            if not speech:
                return ""

            # Handle multiple matching key phrases. This appears to be a
            # quirk of how Pocket Sphinx 'kws' searches work. Get the best
            # match instead if this is the case.
            recognised_phrases = speech.split("  ")
            if len(recognised_phrases) > 1:
                # Remove trailing space from the last phrase.
                recognised_phrases[len(recognised_phrases) - 1].rstrip()
                speech = self._get_best_hypothesis(recognised_phrases)
            else:
                speech = speech.rstrip()  # remove trailing whitespace.

        # Notify observers if a keyphrase was matched.
        results_obj = None  # TODO Use PS results object once implemented
        result = speech if speech in self._keyphrase_functions else ""
        words = tuple(result.split())
        if words:
            self._recognition_observer_manager.notify_recognition(
                words, None, None, results_obj
            )

        # Call the registered function if there was a match and the function
        # is callable.
        func = self._keyphrase_functions.get(speech, None)
        if callable(func):
            try:
                func()
            except Exception as e:
                self._log.exception(
                    "Exception caught when executing the function for "
                    "keyphrase '%s': %s" % (speech, e)
                )

        # Notify observers after calling the keyphrase function.
        if words:
            self._recognition_observer_manager.notify_post_recognition(
                words, None, None, results_obj
            )

        return result

    @classmethod
    def _generate_words_rules(cls, words, mimicking, all_dictation):
        # Convert words to Unicode, treat all uppercase words as dictation
        # words and other words as grammar words.
        # Minor note: this won't work for languages without capitalisation.
        result = []
        for word in words.split():
            if isinstance(word, binary_type):
                word = word.decode(locale.getpreferredencoding())
            if all_dictation or word.isupper() and mimicking:
                # Convert dictation words to lowercase for consistent
                # output.
                result.append((word.lower(), 1000000))
            else:
                result.append((word, 0))
        return tuple(result)

    def _process_hypotheses(self, speech, mimicking):
        """
        Internal method to process speech hypotheses. This should only be called
        from 'SphinxEngine._hypothesis_callback' because that method does important
        post processing.

        :param speech: speech
        :param mimicking: whether to treat speech as mimicked speech.
        :rtype: tuple
        """
        # Check key phrases search first.
        keyphrase = self._process_key_phrases(speech, mimicking)
        if keyphrase:
            # Keyphrase search matched.
            return True, keyphrase

        # Otherwise do grammar processing.
        processing_occurred = False
        hypotheses = {}
        wrappers = self._grammar_wrappers.copy().values()

        # Save the LM hypothesis separately because it will almost always be favoured
        # over grammar hypotheses.
        lm_hypothesis = speech

        # Count exclusive grammars.
        exclusive_count = 0
        for wrapper in wrappers:
            if wrapper.exclusive:
                exclusive_count += 1

        # Collect each active grammar wrapper.
        # Only include exclusive grammars if at least one is loaded.
        if exclusive_count:
            wrappers = [w for w in wrappers
                        if w.exclusive and w.grammar_active]
        else:
            wrappers = [w for w in wrappers if w.grammar_active]

        # No grammar has been loaded.
        if not wrappers:
            return processing_occurred, speech

        # Batch process audio buffers for each active grammar. Store each
        # hypothesis.
        for wrapper in wrappers:
            if mimicking:
                # Just use 'speech' for everything if mimicking.
                hyp = speech
            else:
                # Switch to the search for this grammar and re-process the
                # audio.
                self._set_grammar(wrapper, True)
                hyp = self._decoder.batch_process(
                    self._audio_buffers,
                    use_callbacks=False
                )
                if hyp:
                    hyp = hyp.hypstr

            # Set the hypothesis in the dictionary.
            hypotheses[wrapper.search_name] = hyp

        # Get the best hypothesis.
        speech = self._get_best_hypothesis(list(hypotheses.values()))
        if not speech and not lm_hypothesis:
            return processing_occurred, speech

        if speech:
            # Process speech using the first matching grammar.
            words_rules = self._generate_words_rules(speech, mimicking, False)
            for wrapper in wrappers:
                if hypotheses[wrapper.search_name] != speech:
                    continue

                processing_occurred = wrapper.process_words(words_rules)
                if processing_occurred:
                    break

        if not processing_occurred:
            # Process grammars using the LM hypothesis as dictation words.
            dictation_words = self._generate_words_rules(lm_hypothesis, mimicking,
                                                         True)
            for wrapper in wrappers:
                processing_occurred = wrapper.process_words(dictation_words)
                if processing_occurred:
                    break

        # Return whether processing occurred and the final speech hypothesis for
        # post processing.
        return processing_occurred, speech

    def process_buffer(self, buf):
        """
        Recognise speech from an audio buffer.

        This method is meant to be called in sequence for multiple audio
        buffers. It will do nothing if :meth:`connect` hasn't been called.

        :param buf: audio buffer
        :type buf: str
        """
        if not self._decoder:
            return

        # Cancel current recognition if it has been requested.
        if self._cancel_recognition_next_time:
            self._decoder.end_utterance()
            self._audio_buffers = []
            self._cancel_recognition_next_time = False

        # Keep a list of buffers for possible reprocessing using different Pocket
        # Sphinx searches later.
        self._audio_buffers.append(buf)

        # Call the timer callback if it is set.
        self.call_timer_callback()

        # Process audio.
        try:
            self._recognising = True
            self._decoder.process_audio(buf)
        finally:
            self._recognising = False

    def process_wave_file(self, path):
        """
        Recognise speech from a wave file and return the recognition results.

        This method checks that the wave file is valid. It raises an error
        if the file doesn't exist, if it can't be read or if the WAV header
        values do not match those in the engine configuration.

        If recognition is paused (sleep mode), this method will call
        :meth:`resume_recognition`.

        The wave file must use the same sample width, sample rate and number
        of channels that the acoustic model uses.

        If the file is valid, :meth:`process_buffer` is then used to process
        the audio.

        Multiple utterances are supported.

        :param path: wave file path
        :raises: IOError | OSError | ValueError
        :returns: recognition results
        :rtype: generator
        """
        if not self._decoder:
            self.connect()

        # This method's implementation has been adapted from the PyAudio
        # play wave example:
        # http://people.csail.mit.edu/hubert/pyaudio/#play-wave-example

        # Check that path is a valid file.
        if not os.path.isfile(path):
            raise IOError("'%s' is not a file. Please use a different file path.")

        # Get required audio configuration from the engine config.
        channels, sample_width, rate, chunk = (
            self.config.CHANNELS,
            self.config.SAMPLE_WIDTH,
            self.config.RATE,
            self.config.FRAMES_PER_BUFFER
        )

        # Make sure recognition is not paused.
        if self.recognition_paused:
            self.resume_recognition(notify=False)

        # Open the wave file. Use contextlib to make sure that the file is
        # closed whether errors are raised or not.
        # Also register a custom recognition observer for the duration.
        obs = WaveRecognitionObserver(self)
        with contextlib.closing(wave.open(path, "rb")) as wf, obs as obs:
            # Validate the wave file's header.
            if wf.getnchannels() != channels:
                message = ("WAV file '%s' should use %d channel(s), not %d!"
                           % (path, channels, wf.getnchannels()))
            elif wf.getsampwidth() != sample_width:
                message = ("WAV file '%s' should use sample width %d, not "
                           "%d!" % (path, sample_width, wf.getsampwidth()))
            elif wf.getframerate() != rate:
                message = ("WAV file '%s' should use sample rate %d, not "
                           "%d!" % (path, rate, wf.getframerate()))
            else:
                message = None

            if message:
                raise ValueError(message)

            # Use process_buffer to process each buffer.
            for _ in range(0, int(wf.getnframes() / chunk) + 1):
                data = wf.readframes(chunk)
                if not data:
                    break

                self.process_buffer(data)

                # Get the results from the observer.
                if obs.words:
                    yield obs.words
                    obs.words = ""

        # Log warnings if speech start or end weren't detected.
        if not obs.complete:
            self._log.warning("Speech start/end wasn't detected in the wave "
                              "file!")
            self._log.warning("Perhaps the Sphinx '-vad_prespeech' value "
                              "should be higher?")
            self._log.warning("Or maybe '-vad_startspeech' or "
                              "'-vad_postspeech' should be lower?")

    def _do_recognition(self):
        """
        Start recognising from the default recording device until
        :meth:`disconnect` is called.

        Recognition can be paused and resumed using either the sleep/wake
        key phrases or by calling :meth:`pause_recognition` or
        :meth:`resume_recognition`.

        To configure audio input settings, modify the engine's ``CHANNELS``,
        ``RATE``, ``SAMPLE_WIDTH`` and/or ``FRAMES_PER_BUFFER``
        configuration options.
        """
        if not self._decoder:
            self.connect()

        # Start recognising in a loop.
        self._recorder.start()
        self._cancel_recognition_next_time = False
        while self.recognising:
            for buf in self._recorder.get_buffers():
                self.process_buffer(buf)

        # Free engine resources after recognition has stopped.
        self._free_engine_resources()

    def mimic(self, words):
        """ Mimic a recognition of the given *words* """
        if isinstance(words, (list, tuple)):
            words = " ".join(words)

        # Fail on empty input.
        if not words:
            raise MimicFailure("Invalid mimic input %r" % words)

        if self.recognition_paused and words == self.config.WAKE_PHRASE:
            self.resume_recognition()
            return

        # Pretend that Sphinx has started processing speech
        self._speech_start_callback(True)

        # Process the words as if they were spoken
        result = self._hypothesis_callback(words, True)
        if not result:
            raise MimicFailure("No matching rule found for words %s."
                               % words)

    def mimic_phrases(self, *phrases):
        """
        Mimic a recognition of the given *phrases*.

        This method accepts variable phrases instead of a list of words.
        """
        # Pretend that Sphinx has started processing speech
        self._speech_start_callback(True)

        # Process phrases as if they were spoken
        wake_phrase = self.config.WAKE_PHRASE
        for phrase in phrases:
            if self.recognition_paused and phrase == wake_phrase:
                self.resume_recognition()
                continue

            result = self._hypothesis_callback(phrase, True)
            if not result:
                raise MimicFailure("No matching rule found for words %s."
                                   % phrase)

    def speak(self, text):
        """"""
        self._log.warning("text-to-speech is not implemented for this "
                          "engine.")
        self._log.warning("Printing text instead.")
        print(text)

    def _get_language(self):
        return self.config.LANGUAGE

    def _has_quoted_words_support(self):
        return False

    # ----------------------------------------------------------------------
    # Training-related methods

    def write_transcript_files(self, fileids_path, transcription_path):
        """
        Write .fileids and .transcription files for files in the training
        data directory and write them to the specified file paths.

        This method will raise an error if the ``TRAINING_DATA_DIR``
        configuration option is not set to an existing directory.

        :param fileids_path: path to .fileids file to create.
        :param transcription_path: path to .transcription file to create.
        :type fileids_path: str
        :type transcription_path: str
        :raises: IOError | OSError
        """
        write_transcript_files(
            self.config, fileids_path, transcription_path
        )

    @property
    def training_session_active(self):
        """
        Whether a training session is in progress.

        :rtype: bool
        """
        return self._training_session_active

    def start_training_session(self):
        """
        Start the training session. This will stop recognition processing
        until either :meth:`end_training_session` is called or the end
        training keyphrase is heard.
        """
        data_dir = self.config.TRAINING_DATA_DIR
        if not data_dir or not os.path.isdir(data_dir):
            self._log.warning("Training data will not be recorded; '%s' is "
                              "not a directory" % data_dir)

        if not self._training_session_active:
            self._log.info("Training session has started. No rule "
                           "actions will be processed. ")
            self._log.info("Say '%s' to end the session."
                           % self.config.END_TRAINING_PHRASE)
            self._training_session_active = True

    def end_training_session(self):
        """
        End the training if one is in progress. This will allow recognition
        processing once again.
        """
        if self._training_session_active:
            self._log.info("Ending training session.")
            self._log.info("Rule actions will now be processed normally "
                           "again.")
            self._training_session_active = False

    # ----------------------------------------------------------------------
    # Recognition loop control methods
    # Stopping recognition loop is done using disconnect()

    @property
    def recognition_paused(self):
        """
        Whether the engine is waiting for the wake phrase to be heard or for
        :meth:`resume_recognition` to be called.

        :rtype: bool
        """
        return self._recognition_paused

    def pause_recognition(self):
        """
        Pause recognition and wait for :meth:`resume_recognition` to be
        called or for the wake keyphrase to be spoken.
        """
        if not self._decoder:
            return

        self._recognition_paused = True

        # Switch to the wake keyphrase search if a wake keyphrase has been
        # set.
        self._set_default_search()
        if not self.config.WAKE_PHRASE:
            self._log.warning("No wake phrase has been set.")
            self._log.warning("Use engine.resume_recognition() to wake up.")

        # Define temporary callback for the decoder.
        def hypothesis(hyp):
            # Clear any recorded audio buffers.
            self._recorder.clear_buffers()
            s = hyp.hypstr if hyp else None

            # Resume recognition if s is the wake keyphrase.
            if s and s.strip() == self.config.WAKE_PHRASE.strip():
                self.resume_recognition()
            elif self.config.WAKE_PHRASE:
                self._log.debug("Didn't hear %s" % self.config.WAKE_PHRASE)

            # Clear audio buffers
            self._audio_buffers = []

        # Override decoder hypothesis callback.
        self._decoder.hypothesis_callback = hypothesis

    def resume_recognition(self, notify=True):
        """
        Resume listening for grammar rules and key phrases.
        """
        if not self._decoder:
            return

        self._recognition_paused = False

        # Notify observers about recognition resume.
        keyphrase = self.config.WAKE_PHRASE
        words = tuple(keyphrase.strip().split())
        results_obj = None  # TODO Use PS results object once implemented
        if words and notify:
            manager = self._recognition_observer_manager
            arguments = (words, None, None, results_obj)
            manager.notify_recognition(*arguments)
            manager.notify_post_recognition(*arguments)

        # Restore the callbacks to normal
        def hypothesis(hyp):
            # Set default search result.
            self._default_search_result = hyp

            # Set speech to the hypothesis string or None if there isn't one
            speech = hyp.hypstr if hyp else None
            return self._hypothesis_callback(speech, False)

        self._decoder.hypothesis_callback = hypothesis

        # Switch to the default search.
        self._set_default_search()

    def cancel_recognition(self):
        """
        If a recognition was in progress, cancel it before processing the
        next audio buffer.
        """
        self._cancel_recognition_next_time = True



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This module is a simple example of Dragonfly use.

    It shows how to use Dragonfly's Grammar, AppContext, and MappingRule
    classes.  This module can be activated in the same way as other
    Natlink macros by placing it in the "NatLink/MacroSystem" folder.
"""

from dragonfly import (Grammar, AppContext, MappingRule, Choice,
                       Key, Text, Pause)


#---------------------------------------------------------------------------
# Create this module's grammar and the context under which it'll be active.

grammar_context = AppContext(executable="outlook")
grammar = Grammar("outlook_example", context=grammar_context)


#---------------------------------------------------------------------------
# Create a rule for static keyboard shortcuts.

static_rule = MappingRule(
    name="static",
    mapping={
             "new folder":                   Key("cs-e"),
             "new email":                    Key("cs-m"),
             "new appointment":              Key("cs-a"),
             "new meeting request":          Key("cs-q"),
             "new contact":                  Key("cs-c"),
             "new distribution list":        Key("cs-l"),
             "new note":                     Key("cs-n"),
             "new task":                     Key("cs-k"),

             "accept and edit":              Key("a-a, c, e, enter"),
             "accept and send":              Key("a-a, c, s, enter"),
             "accept without response":      Key("a-a, c, d, enter"),
             "tentative and edit":           Key("a-a, a, e, enter"),
             "tentative and send":           Key("a-a, a, s, enter"),
             "tentative without response":   Key("a-a, a, d, enter"),

             "folder":                       Key("c-y"),
            },
    )

# Add the action rule to the grammar instance.
grammar.add_rule(static_rule)


#---------------------------------------------------------------------------
# Create a rule for our list of folders.

folder_map = {
              "inbox":         "inbox",
              "sent":          "sent items",
              "deleted":       "deleted",
              "outbox":        "outbox",
             }

folder_rule = MappingRule(
    name="folder",
    mapping={
             "folder <folder>":   Key("c-1, home")
                                   + Text("%(folder)s\n")
                                   + Key("space, tab"),
             "move to <folder>":  Key("cs-v, home/10")
                                   + Text("%(folder)s")
                                   + Pause("75")
                                   + Key("enter"),
            },
    extras=[
            Choice("folder", folder_map),
           ],
    )

# Add the action rule to the grammar instance.
grammar.add_rule(folder_rule)


#---------------------------------------------------------------------------
# Load the grammar instance and define how to unload it.

grammar.load()

# Unload function which will be called by natlink at unload time.
def unload():
    global grammar
    if grammar: grammar.unload()
    grammar = None



#
# This file is based on how Aenea handles and generates security tokens.
# The original copyright notice is reproduced below.
#
# This file is part of Aenea
#
# Aenea is free software: you can redistribute it and/or modify it under
# the terms of version 3 of the GNU Lesser General Public License as
# published by the Free Software Foundation.
#
# Aenea is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Aenea.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (2014) Alex Roper
# Alex Roper <alex@aroper.net>

import base64
import os


def compare_security_token(expected, actual):
    if len(expected) != len(actual):
        return False
    result = 0
    for x, y in zip(expected, actual):
        result |= ord(x) ^ ord(y)
    return result == 0


def generate_security_token():
    return base64.b64encode(os.urandom(32)).decode()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Base Window class
============================================================================

"""

# pylint: disable=W0622,R0904
# Suppress warnings about redefining the built-in 'id' function and too many
# public methods.

from locale import getpreferredencoding

from six import string_types, integer_types, binary_type

from .monitor import monitors
from .rectangle import unit
from .window_movers import window_movers

#===========================================================================


class BaseWindow(object):
    """
        The base Window class for controlling and placing windows.

    """

    #-----------------------------------------------------------------------
    # Class attributes to retrieve existing Window objects.

    _windows_by_name = {}
    _windows_by_id = {}

    #-----------------------------------------------------------------------
    # Class methods to create new Window objects.

    @classmethod
    def get_window(cls, id):
        """
        Get a :class:`Window` object given a window id.

        Given the same id, this method will return the same object.
        """
        if id in cls._windows_by_id:
            window = cls._windows_by_id[id]
        else:
            window = cls(id)
            cls._windows_by_id[id] = window
        return window

    @classmethod
    def get_foreground(cls):
        """ Get the foreground window. """
        raise NotImplementedError()

    @classmethod
    def get_matching_windows(cls, executable=None, title=None):
        """
        Find windows with a matching executable or title.

        Window searches are case-insensitive.

        If neither parameter is be specified, then it is effectively the
        same as calling :meth:`get_all_windows`.

        :param executable: -- part of the filename of the application's
           executable to which the target window belongs; not case
           sensitive.
        :param title: -- part of the title of the target window; not case
           sensitive.
        :type executable: str
        :type title: str
        :rtype: list
        """
        # Make window searches case-insensitive.
        if executable:
            executable = executable.lower()
        if title:
            title = title.lower()

        matching = []
        for window in cls.get_all_windows():
            if executable:
                if window.executable.lower().find(executable) == -1:
                    continue
            if title:
                if window.title.lower().find(title) == -1:
                    continue
            matching.append(window)
        return matching

    @classmethod
    def get_all_windows(cls):
        """ Get a list of all windows. """
        raise NotImplementedError()

    #-----------------------------------------------------------------------
    # Methods for initialization and introspection.

    def __init__(self, id):
        self._id = None
        self.id = id
        self._names = set()

    def __repr__(self):
        args = list(self._names)
        return "%s(%s)" % (self.__class__.__name__, ", ".join(args))

    #-----------------------------------------------------------------------
    # Methods that control attribute access.

    @property
    def id(self):
        """ Protected access to id attribute. """
        return self._id

    @id.setter
    def id(self, value):
        self._set_id(value)

    def _set_id(self, id):
        if not isinstance(id, integer_types):
            raise TypeError("Window id/handle must be integer or long,"
                            " but received {0!r}".format(id))
        self._id = id
        self._windows_by_id[id] = self

    # The 'handle' and '_handle' attributes have been kept in for backwards-
    # compatibility. They just reference the BaseWindow 'id' property.

    handle = property(fget=lambda self: self._id,
                      fset=_set_id,
                      doc="Protected access to handle attribute.")
    _handle = property(fget=lambda self: self._id)

    def _get_name(self):
        if not self._names:
            return None
        for name in self._names:
            return name

    def _set_name(self, name):
        assert isinstance(name, string_types)
        self._names.add(name)
        self._windows_by_name[name] = self

    name = property(fget=_get_name,
                    fset=_set_name,
                    doc="Protected access to name attribute.")

    #-----------------------------------------------------------------------
    # Methods and properties for window attributes.

    def _get_window_text(self):
        # Method to get the window title.
        raise NotImplementedError()

    def _get_class_name(self):
        # Method to get the window class name.
        raise NotImplementedError()

    def _get_window_module(self):
        # Method to get the window executable.
        raise NotImplementedError()

    def _get_window_pid(self):
        # Method to get the window process ID.
        raise NotImplementedError()

    @property
    def title(self):
        """ Read-only access to the window's title. """
        window_text = self._get_window_text()
        # PY2
        if isinstance(window_text, binary_type):
            return window_text.decode(getpreferredencoding())
        else:
            return window_text

    @property
    def classname(self):
        """ Read-only access to the window's class name. """
        return self._get_class_name()

    #: Alias of :attr:`classname`.
    cls_name = classname

    @property
    def executable(self):
        """ Read-only access to the window's executable. """
        return self._get_window_module()

    @property
    def pid(self):
        """
        Read-only access to the window's process ID.

        This will be the PID of the window's process, not any subprocess.

        If the window has no associated process id, this will return
        ``None``.

        :returns: pid
        :rtype: int | None
        """
        return self._get_window_pid()

    @property
    def is_minimized(self):
        """ Whether the window is currently minimized. """
        raise NotImplementedError()

    @property
    def is_maximized(self):
        """ Whether the window is currently maximized. """
        raise NotImplementedError()

    @property
    def is_visible(self):
        """
        Whether the window is currently visible.

        This may be indeterminable for some windows.
        """
        raise NotImplementedError()

    def matches(self, context):
        """ Whether the window matches the given context. """
        return context.matches(self.executable, self.title, self.handle)

    #-----------------------------------------------------------------------
    # Methods related to window geometry.

    def get_position(self):
        """
        Method to get the window's position as a :class:`Rectangle` object.

        :returns: window position
        :rtype: Rectangle
        """
        raise NotImplementedError()

    def set_position(self, rectangle):
        """
        Method to set the window's position using a :class:`Rectangle`
        object.

        :param rectangle: window position
        :type rectangle: Rectangle
        """
        raise NotImplementedError()

    def get_containing_monitor(self):
        """
        Method to get the :class:`Monitor` containing the window.

        This checks which monitor contains the center of the window.

        :returns: containing monitor
        :rtype: Monitor
        """
        center = self.get_position().center
        for monitor in monitors:
            if monitor.rectangle.contains(center):
                return monitor
        # Fall through, return first monitor.
        return monitors[0]

    def get_normalized_position(self):
        """
        Method to get the window's normalized position.

        This is useful when working with multiple monitors.

        :returns: normalized position
        :rtype: Rectangle
        """
        monitor = self.get_containing_monitor()
        rectangle = self.get_position()
        rectangle.renormalize(monitor.rectangle, unit)
        return rectangle

    def set_normalized_position(self, rectangle, monitor=None):
        """
        Method to get the window's normalized position.

        This is useful when working with multiple monitors.

        :param rectangle: window position
        :type rectangle: Rectangle
        :param monitor: monitor to normalize to (default: the first one).
        :type monitor: Monitor
        """
        if not monitor:
            monitor = self.get_containing_monitor()

        rectangle.renormalize(unit, monitor.rectangle)
        self.set_position(rectangle)

    #-----------------------------------------------------------------------
    # Methods for miscellaneous window control.

    def minimize(self):
        """ Minimize the window (if possible). """
        raise NotImplementedError()

    def maximize(self):
        """ Maximize the window (if possible). """
        raise NotImplementedError()

    def close(self):
        """ Close the window (if possible). """
        raise NotImplementedError()

    def restore(self):
        """ Restore the window if it is minimized or maximized. """
        raise NotImplementedError()

    def set_foreground(self):
        """ Set the window as the foreground (active) window. """
        raise NotImplementedError()

    def set_focus(self):
        """
        Set the window as the active window without raising it.

        *Note*: this method will behave like :meth:`set_foreground()` in
        environments where this isn't possible.
        """
        raise NotImplementedError()

    def move(self, rectangle, animate=None):
        """
        Move the window, optionally animating its movement.

        :param rectangle: new window position and size
        :param animate: name of window mover
        :type rectangle: Rectangle
        :type animate: str
        """
        if not animate:
            self.set_position(rectangle)
        else:
            try:
                window_mover = window_movers[animate]
            except KeyError:
                # If the given window mover name isn't found, don't animate.
                self.set_position(rectangle)
            else:
                window_mover.move_window(self, self.get_position(),
                                         rectangle)



#
# This file is part of Dragonfly.
# (c) Copyright 2019 by David Zurow
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Compiler classes for Kaldi backend
"""

import collections, types

from .dictation                 import AlternativeDictation, DefaultDictation, UserDictation
from ..base                     import CompilerBase, CompilerError
from ...grammar                 import elements as elements_

from kaldi_active_grammar import WFST, KaldiRule
from kaldi_active_grammar import Compiler as KaldiAGCompiler

import six
from six import text_type
from six.moves import map, range


#---------------------------------------------------------------------------
# Utilities

_trace_level=0
def trace_compile(func):
    return func
    def dec(self, element, src_state, dst_state, grammar, fst):
        global _trace_level
        s = '%s %s: compiling %s' % (grammar.name, '==='*_trace_level, element)
        l = 140-len(s)
        s += ' '*l + '| %-20s %s -> %s' % (id(fst), src_state, dst_state)
        grammar._log_load.error(s)
        _trace_level+=1
        ret = func(self, element, src_state, dst_state, grammar, fst)
        _trace_level-=1
        grammar._log_load.error('%s %s: compiling %s.' % (grammar.name, '...'*_trace_level, element))
        return ret
    return dec

InternalGrammar = collections.namedtuple('InternalGrammar', 'name')
InternalRule = collections.namedtuple('InternalRule', 'name gstring')

MockLiteral = collections.namedtuple('MockLiteral', 'words')


#---------------------------------------------------------------------------

class KaldiCompiler(CompilerBase, KaldiAGCompiler):

    def __init__(self, model_dir, tmp_dir, auto_add_to_user_lexicon=None, lazy_compilation=None, **kwargs):
        CompilerBase.__init__(self)
        KaldiAGCompiler.__init__(self, model_dir=model_dir, tmp_dir=tmp_dir, **kwargs)

        self.auto_add_to_user_lexicon = bool(auto_add_to_user_lexicon)
        self.lazy_compilation = bool(lazy_compilation)

        self.kaldi_rule_by_rule_dict = collections.OrderedDict()  # Rule -> KaldiRule
        # self._grammar_rule_states_dict = dict()  # FIXME: disabled!
        self.kaldi_rules_by_listreflist_dict = collections.defaultdict(set)  # Rule -> Set[KaldiRule]
        self.internal_grammar = InternalGrammar('!kaldi_engine_internal')

    impossible_word = property(lambda self: self._longest_word.lower())  # FIXME
    unknown_word = property(lambda self: self._oov_word)

    #-----------------------------------------------------------------------
    # Methods for handling lexicon translation.

    # FIXME: documentation
    translation_dict = {
    }

    # FIXME: documentation
    untranslation_dict = { v: k for k, v in translation_dict.items() }
    translation_dict.update({
    })

    def untranslate_output(self, output):
        for old, new in six.iteritems(self.untranslation_dict):
            output = output.replace(old, new)
        return output

    def translate_words(self, words):
        # Unused
        if self.translation_dict:
            new_words = []
            for word in words:
                for old, new in six.iteritems(self.translation_dict):
                    word = word.replace(old, new)
                new_words.extend(word.split())
            words = new_words
        words = [word.lower() for word in words]
        for i in range(len(words)):
            if words[i] not in self.lexicon_words:
                words[i] = self.handle_oov_word(words[i])
        return words

    def handle_oov_word(self, word):
        if self.auto_add_to_user_lexicon:
            try:
                pronunciations = self.add_word(word, lazy_compilation=True)
            except Exception as e:
                self._log.exception("%s: exception automatically adding word %r" % (self, word))
            else:
                for phones in pronunciations:
                    self._log.warning("%s: Word not in lexicon (generated automatic pronunciation): %r [%s]" % (self, word, ' '.join(phones)))
                return word

        self._log.warning("%s: Word %r not in lexicon (will NOT be recognized; see documentation about user lexicon and auto_add_to_user_lexicon)" % (self, word))
        word = self.impossible_word
        return word

    #-----------------------------------------------------------------------
    # Methods for compiling grammars.

    def compile_grammar(self, grammar, engine):
        self._log.debug("%s: Compiling grammar %s." % (self, grammar.name))

        kaldi_rule_by_rule_dict = collections.OrderedDict()  # Rule -> KaldiRule
        for rule in grammar.rules:
            if rule.exported:
                if rule.element is None:
                    # We cannot deal with an empty rule (could be fixed by refactoring)
                    raise CompilerError("Invalid None element for %s in %s" % (rule, grammar))

                kaldi_rule = KaldiRule(self,
                    name='%s::%s' % (grammar.name, rule.name),
                    has_dictation=None)  # has_dictation is set to True during compilation below if that is the case
                kaldi_rule.parent_grammar = grammar
                kaldi_rule.parent_rule = rule
                kaldi_rule_by_rule_dict[rule] = kaldi_rule

                try:
                    self._compile_rule_root(rule, grammar, kaldi_rule)
                    kaldi_rule.has_dictation = bool(kaldi_rule.has_dictation)  # Convert None to False
                except Exception:
                    raise self.make_compiler_error_for_kaldi_rule(kaldi_rule)

        self.kaldi_rule_by_rule_dict.update(kaldi_rule_by_rule_dict)
        return kaldi_rule_by_rule_dict

    def _compile_rule_root(self, rule, grammar, kaldi_rule):
        src_state, dst_state = self._compile_rule(rule, grammar, kaldi_rule, kaldi_rule.fst, export=True)
        if kaldi_rule.fst.native and not kaldi_rule.fst.has_path():
            # Impossible paths break AGF compilation, so bolt on an Impossible element. This is less than ideal, but what are you doing compiling this anyway?
            self._compile_impossible(None, src_state, dst_state, grammar, kaldi_rule, kaldi_rule.fst)
        kaldi_rule.compile(lazy=self.lazy_compilation)

    def _compile_rule(self, rule, grammar, kaldi_rule, fst, export):
        """ :param export: whether rule is exported (a root rule) """
        # Determine whether this rule has already been compiled.
        # if (grammar.name, rule.name) in self._grammar_rule_states_dict:
        #     self._log.debug("%s: Already compiled rule %s%s." % (self, rule.name, ' [EXPORTED]' if export else ''))
        #     return self._grammar_rule_states_dict[(grammar.name, rule.name)]
        # else:
        self._log.debug("%s: Compiling rule %s%s." % (self, rule.name, ' [EXPORTED]' if export else ''))

        if export:
            # Root rule, so must handle grammar's weight, in addition to this rule's weight
            weight = self.get_weight(grammar) * self.get_weight(rule)
            outer_src_state = fst.add_state(initial=True)
            inner_src_state = fst.add_state()
            fst.add_arc(outer_src_state, inner_src_state, None, weight=weight)
            dst_state = fst.add_state(final=True)

        else:
            # Only handle this rule's weight
            weight = self.get_weight(rule)
            outer_src_state = fst.add_state()
            inner_src_state = fst.add_state()
            fst.add_arc(outer_src_state, inner_src_state, None, weight=weight)
            dst_state = fst.add_state()

        self.compile_element(rule.element, inner_src_state, dst_state, grammar, kaldi_rule, fst)
        # self._grammar_rule_states_dict[(grammar.name, rule.name)] = (src_state, dst_state)
        return (outer_src_state, dst_state)

    def unload_grammar(self, grammar, rules, engine):
        for rule in rules:
            kaldi_rule = self.kaldi_rule_by_rule_dict[rule]
            # Unload kaldi_rule: destroy() handles KaldiAGCompiler stuff; we must handle ours
            kaldi_rule.destroy()
            del self.kaldi_rule_by_rule_dict[rule]
            for kaldi_rules_set in self.kaldi_rules_by_listreflist_dict.values():
                kaldi_rules_set.discard(kaldi_rule)
            # NOTE: the kaldi_rule_by_rule_dict we returned from compile_grammar() is not updated, but it should be dropped upon unload anyway!

    def update_list(self, lst, grammar):
        # Note: we update all rules in all grammars that reference this list (unlike WSR/natlink?)
        lst_kaldi_rules = self.kaldi_rules_by_listreflist_dict[id(lst)]
        for kaldi_rule in lst_kaldi_rules:
            with kaldi_rule.reload():
                self._compile_rule_root(kaldi_rule.parent_rule, grammar, kaldi_rule)

    #-----------------------------------------------------------------------
    # Methods for compiling elements.

    _eps_like_nonterms = frozenset()  # Dictation is non-empty now ('#nonterm:dictation', '#nonterm:dictation_cloud')

    def compile_element(self, element, *args, **kwargs):
        """Compile element in FST (from src_state to dst_state) and return result."""
        # Look for a compiler method to handle the given element.
        for element_type, compiler in self.element_compilers:
            if isinstance(element, element_type):
                return compiler(self, element, *args, **kwargs)
        # Didn't find a compiler method for this element type.
        raise NotImplementedError("Compiler %s not implemented for element type %s." % (self, element))

    # @trace_compile
    def _compile_sequence(self, element, src_state, dst_state, grammar, kaldi_rule, fst):
        src_state = self.add_weight_linkage(src_state, dst_state, self.get_weight(element), fst)
        children = element.children
        # Optimize for special lengths
        if len(children) == 0:
            fst.add_arc(src_state, dst_state, None)
            return

        elif len(children) == 1:
            return self.compile_element(children[0], src_state, dst_state, grammar, kaldi_rule, fst)

        else:  # len(children) >= 2:
            # Handle Repetition elements differently as a special case
            is_repetition = isinstance(element, elements_.Repetition)
            if is_repetition and element.optimize:
                # Repetition...
                # Insert new states, so back arc only affects child
                s1 = fst.add_state()
                s2 = fst.add_state()
                fst.add_arc(src_state, s1, None)
                # NOTE: to avoid creating an un-decodable epsilon loop, we must not allow an all-epsilon child here (compile_graph_agf should check this)
                self.compile_element(children[0], s1, s2, grammar, kaldi_rule, fst)
                if not fst.has_eps_path(s1, s2, self._eps_like_nonterms):
                    fst.add_arc(s2, s1, fst.eps_disambig, fst.eps)  # Back arc, uses eps_disambig ('#0')
                    fst.add_arc(s2, dst_state, None)
                    return

                else:
                    # Cannot do optimize path, because of epsilon loop, so finish up with Sequence path
                    self._log.warning("%s: Cannot optimize Repetition element, because its child element can match empty string;"
                        " falling back to inefficient non-optimize path. (this is not that bad)" % self)
                    states = [src_state, s2] + [fst.add_state() for i in range(len(children)-2)] + [dst_state]
                    for i, child in enumerate(children[1:], start=1):
                        s1 = states[i]
                        s2 = states[i + 1]
                        self.compile_element(child, s1, s2, grammar, kaldi_rule, fst)
                    return

            else:
                # Sequence, not Repetition...
                # Insert new states for individual children elements
                states = [src_state] + [fst.add_state() for i in range(len(children)-1)] + [dst_state]
                for i, child in enumerate(children):
                    s1 = states[i]
                    s2 = states[i + 1]
                    self.compile_element(child, s1, s2, grammar, kaldi_rule, fst)
                return

    # @trace_compile
    def _compile_alternative(self, element, src_state, dst_state, grammar, kaldi_rule, fst):
        src_state = self.add_weight_linkage(src_state, dst_state, self.get_weight(element), fst)
        for child in element.children:
            self.compile_element(child, src_state, dst_state, grammar, kaldi_rule, fst)

    # @trace_compile
    def _compile_optional(self, element, src_state, dst_state, grammar, kaldi_rule, fst):
        src_state = self.add_weight_linkage(src_state, dst_state, self.get_weight(element), fst)
        self.compile_element(element.children[0], src_state, dst_state, grammar, kaldi_rule, fst)
        fst.add_arc(src_state, dst_state, None)

    # @trace_compile
    def _compile_literal(self, element, src_state, dst_state, grammar, kaldi_rule, fst):
        weight = self.get_weight(element)  # Handle weight internally below, without adding a state
        words = element.words
        words = list(map(text_type, words))
        # words = self.translate_words(words)

        # Special case optimize single-word literal
        if len(words) == 1:
            word = words[0].lower()
            if word not in self.lexicon_words:
                word = self.handle_oov_word(word)
            fst.add_arc(src_state, dst_state, word, weight=weight)

        else:
            words = [word.lower() for word in words]
            for i in range(len(words)):
                if words[i] not in self.lexicon_words:
                    words[i] = self.handle_oov_word(words[i])
            # "Insert" new states for individual words
            states = [src_state] + [fst.add_state() for i in range(len(words)-1)] + [dst_state]
            for i, word in enumerate(words):
                fst.add_arc(states[i], states[i + 1], word, weight=weight)
                weight = None  # Only need to set weight on first arc

    # @trace_compile
    def _compile_rule_ref(self, element, src_state, dst_state, grammar, kaldi_rule, fst):
        weight = self.get_weight(element)  # Handle weight internally below without adding a state
        # Compile target rule "inline"
        rule_src_state, rule_dst_state = self._compile_rule(element.rule, grammar, kaldi_rule, fst, export=False)
        fst.add_arc(src_state, rule_src_state, None, weight=weight)
        fst.add_arc(rule_dst_state, dst_state, None)

    # @trace_compile
    def _compile_list_ref(self, element, src_state, dst_state, grammar, kaldi_rule, fst):
        src_state = self.add_weight_linkage(src_state, dst_state, self.get_weight(element), fst)
        # list_rule_name = "__list_%s" % element.list.name
        if element.list not in grammar.lists:
            # Should only happen during initial compilation; during updates, we must skip this
            grammar.add_list(element.list)
        self.kaldi_rules_by_listreflist_dict[id(element.list)].add(kaldi_rule)
        for child_str in element.list.get_list_items():
            self._compile_literal(MockLiteral(child_str.split()), src_state, dst_state, grammar, kaldi_rule, fst)

    # @trace_compile
    def _compile_dictation(self, element, src_state, dst_state, grammar, kaldi_rule, fst):
        kaldi_rule.has_dictation = True
        src_state = self.add_weight_linkage(src_state, dst_state, self.get_weight(element), fst)
        # fst.add_arc(src_state, dst_state, '#nonterm:dictation', olabel=WFST.eps)
        extra_state = fst.add_state()
        cloud_dictation = isinstance(element, (AlternativeDictation, DefaultDictation)) and element.cloud
        dictation_nonterm = '#nonterm:dictation_cloud' if cloud_dictation else '#nonterm:dictation'
        fst.add_arc(src_state, extra_state, '#nonterm:dictation', dictation_nonterm)
        # Accepts zero or more words
        fst.add_arc(extra_state, dst_state, WFST.eps, '#nonterm:end')
        # fst.add_arc(extra_state, dst_state, '!SIL', '#nonterm:end')  # Causes problems with lack of phones during decoding

    # @trace_compile
    def _compile_impossible(self, element, src_state, dst_state, grammar, kaldi_rule, fst):
        # FIXME: not impossible enough (lower probability?)
        # Note: setting weight=0 breaks compilation!
        fst.add_arc(src_state, dst_state, self.impossible_word, weight=1e-10)

    # @trace_compile
    def _compile_empty(self, element, src_state, dst_state, grammar, kaldi_rule, fst):
        src_state = self.add_weight_linkage(src_state, dst_state, self.get_weight(element), fst)
        fst.add_arc(src_state, dst_state, WFST.eps)

    #-----------------------------------------------------------------------
    # Utility methods.

    def get_weight(self, obj, name='weight'):
        """ Gets the weight of given grammar or rule, checking for invalid values. """
        weight = getattr(obj, name, 1)
        if isinstance(obj, (elements_.Dictation, UserDictation)) and isinstance(weight, types.FunctionType):
            # Ignore crazy string method handling on Dictation elements; use default value
            weight = 1
        try:
            weight = float(weight)
        except TypeError:
            self._log.error("%s: Weight must be a numeric, but %s.%s is %s: %r" % (self, obj, name, type(weight), weight))
            weight = 1
        if weight <= 0:
            self._log.error("%s: Weight cannot be negative or 0, but %s.%s is %s" % (self, obj, name, weight))
            weight = 1e-9
        return weight

    def add_weight_linkage(self, outer_src_state, dst_state, weight, fst):
        """ Returns new source state, to be used by the caller as the effective source state. Only modifies if weight is non-default. """
        if (weight is None) or (weight == 1):
            return outer_src_state
        # self._log.debug("%s: Adding weight linkage for weight=%s" % (self, weight))
        inner_src_state = fst.add_state()
        fst.add_arc(outer_src_state, inner_src_state, None, weight=weight)
        return inner_src_state

    def make_compiler_error_for_kaldi_rule(self, kaldi_rule):
        message = "Exception while compiling %s in %s" % (kaldi_rule.parent_rule, kaldi_rule.parent_grammar)
        if six.PY2: self._log.exception("%s: %s", self, message)  # Imitate PY3's chained exceptions traceback
        rule_tree = kaldi_rule.parent_rule.element.element_tree_string()
        # Limit to at most 100 lines, unless debug logging enabled
        if not self._log.isEnabledFor(10) and rule_tree.count('\n') >= 10:
            rule_tree_lines = rule_tree.split('\n')[:10]
            rule_tree_lines.append("----8<----" * 7)
            rule_tree_lines.append("Printout truncated at 100 lines. To see all, run with:")
            rule_tree_lines.append("    logging.getLogger('kaldi.compiler').setLevel(logging.DEBUG)")
            rule_tree = '\n'.join(rule_tree_lines)
        self._log.error("Failed rule's elements:\n" + rule_tree)
        kaldi_rule.destroy()
        return CompilerError(message)
